# encoding: utf-8

__all__ = ["StockImage", "StockImageException", "TK_IMAGE_FORMATS"]

import logging
import os
import sys
import tkinter as tk
from pathlib import Path
from collections import namedtuple
from dataclasses import dataclass
from enum import Enum
from typing import Any


if sys.version_info < (3, 9):
    import importlib_resources as resources
else:
    import importlib.resources as resources


logger = logging.getLogger(__name__)


class StockImageException(Exception):
    pass


class ImageNotFoundError(StockImageException):
    pass


class ImageFormatNotSupportedError(StockImageException):
    pass


def _iter_package_files(pkg: str):
    subpkg = []
    try:
        for r in resources.files(pkg).iterdir():
            if r.is_file():
                yield r
            if r.is_dir():
                subpkg.append(r.name)
    except NotADirectoryError:
        pass
    for s in subpkg:
        yield from _iter_package_files(f"{pkg}/{s}")


BITMAP_TEMPLATE = "@{0}"
TK_BITMAP_FORMATS = [".xbm"]
TK_PHOTO_FORMATS = [".gif", ".pgm", ".ppm"]


if os.name == "nt":
    TK_BITMAP_FORMATS.append(".ico")
    BITMAP_TEMPLATE = "{0}"

if tk.TkVersion >= 8.6:
    TK_PHOTO_FORMATS.append(".png")


TK_IMAGE_FORMATS = TK_PHOTO_FORMATS + TK_BITMAP_FORMATS


_img_notsupported = """\
R0lGODlhZAAyAIQAAAAAAAsLCxMTExkZGSYmJicnJ11dXYGBgZubm5ycnJ2dnbGxsbOzs8TExMXF
xdXV1dbW1uTk5PLy8v39/f7+/v///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEK
AB8ALAAAAABkADIAAAX+4CeOZGmeaKqubOu+cCzPMmDfeK7vfO//QJ5rQCwaj8ikcslsOpUswGBC
qVqv2Kx2y+16v9nJAKAaFCiVtHrNbrvf8Lh83qYUBigplc7v+/9yFGJkJVJogIiJinAUYyaGi5GS
iI2EJJBuDQ0UChFqmmkNCAqHnAijFKKnnmoRpwgNoagVrqexFaBpEaS0qxWms26VjwOHbaeNDJ4Q
BwgVEAsICQ8SEg8Jp6QIBr5qEAG2z9HTEt+nCxAVCAfpEQzFEQ6nDhGcBga8wo6FxW/IAwISTCAA
rtEDQQMgQChmRR2CKmwWUqmS8FdCiRQenEEQgMCEBAKKIaNwKk1JRvv+LvVz8+/BA4//Qo5RWEyB
GZIBiKTzJmUAqYqNFPZMgObUQJcicw4AZ9IZSksjMB17mLCcw2QQHgigScGdSHYQJKyBIOABhHpA
L5Zt1vRZua8K2TqMM4yfMTb/dl5Ne3YaBAfawIr1tpKTg6wJIixMhW5umsUnIzt9U1fl3TUKSBXQ
m9lOOs+/7ihIY1Pn2DOYbz5DDeFMZm+uR1d4PVs25ZRRV9ZBcxeisd8QfzVkc3n4LzW8ewtPEzz4
bagipE6aTl0f9A/Sq2unXjm3cuTImxtfc2X58vLMxztHL9x3Q/bIcUfX3brU5tA+SRfRy/zOTdqd
+cdaEbYtBJSAaBj+6JMdRLi2H3HyYUcfOJ4EFYFfgHXFQFmD6eKXQo79wwBiipX1FykNoAPNJgOM
eE0E1gigDFbprKPQArcwF6FU5tAT1GLP9APkGheaxYpkWL1IllkZItkiiRZ99mSNTp2k43U81iTQ
RUJ2eRl+RIVIVUis9SSbkwKEVEpaZJJU5WQWYUkfQw/MBOSdupGX0UZvGhQcRoc4idSaUh5U1Jvk
7TgnGjGGdQ0CjQV5WS2QtiMPAj5WRNhd8cyDlqOJRWlRM9pwgykrVxJjzC6ldPKLArC0kk8rr+RY
S4WuyjqpL5zg6uur2aTyCqqp2rXdsdwp+iWyzP7R3Xx7NCutHwiGXSeCatNmS9cdKugBxre7fSvu
uFcMwsIT6KKGnH/otuuuES4EIW+elchr77050ECDdM/q6++/M/AbIcAEF5yCwNYarLDC2P5i7sIQ
K+ztunhEbHHBUlV78cb/YsIgxyDr663GIZcMgw3wmqxyvPmu7HIeCb8sc1Qxz2zzzTjnrPPOPPcs
QwgAOw==
"""


class PitType(Enum):
    PATH = 1
    PACKAGE = 2


@dataclass
class CacheItem:
    def create_tkimage(self):
        ...


@dataclass
class ImgFromData(CacheItem):
    data: Any = None
    format: str = None

    def create_tkimage(self):
        tk.PhotoImage(format=self.format, data=self.data)


@dataclass
class ImgCreated(CacheItem):
    image: Any = None

    def create_tkimage(self):
        return self.image


@dataclass
class ImgFromPath(CacheItem):
    fpath: Any = None

    def _create_with_pillow(self, fpath):
        try:
            from PIL import Image, ImageTk

            aux = Image.open(fpath)
            img = ImageTk.PhotoImage(aux)
            return img
        except ModuleNotFoundError:
            msg = f"Error loading {self.fpath}, image format not supported."
            raise ImageFormatNotSupportedError(msg)

    def create_tkimage(self):
        file_ext = self.fpath.suffix.lower()
        if file_ext in TK_PHOTO_FORMATS:
            img = tk.PhotoImage(file=self.fpath)
        elif file_ext in TK_BITMAP_FORMATS:
            img = tk.BitmapImage(file=self.fpath)
        else:
            img = self._create_with_pillow(self.fpath)
        return img


@dataclass
class ImgFromPackage(ImgFromPath):
    def create_tkimage(self):
        file_ext = Path(str(self.fpath)).suffix.lower()

        with resources.as_file(self.fpath) as file:
            if file_ext in TK_PHOTO_FORMATS:
                img = tk.PhotoImage(file=file)
            elif file_ext in TK_BITMAP_FORMATS:
                img = tk.BitmapImage(file=file)
            else:
                img = self._create_with_pillow(file)
        return img


STOCK_DATA = {"img_not_supported": ImgFromData(_img_notsupported, "gif")}


class StockImage:
    """Maintain references to image name and file.
    When image is used, the class maintains it on memory for tkinter"""

    _stock = STOCK_DATA
    _cached = {}
    _formats = TK_IMAGE_FORMATS
    _resource_pit = {}

    @classmethod
    def clear_cache(cls):
        """Call this before closing tk root"""
        # Prevent tkinter errors on python 2 ??
        for key in cls._cached:
            cls._cached[key] = None
        cls._cached = {}

    @classmethod
    def _register_cache_item(cls, image_id, cache_item):
        if image_id in cls._stock:
            logger.warning("Warning, replacing resource %s", image_id)
        cls._stock[image_id] = cache_item

    @classmethod
    def register(cls, image_id, filename):
        """Register a image file using image_id"""
        fpath = Path(filename) if isinstance(filename, str) else filename
        cls._register_cache_item(image_id, ImgFromPath(fpath))
        logger.info("%s registered as %s", filename, image_id)

    @classmethod
    def register_from_package(cls, image_id, fpath):
        cls._register_cache_item(image_id, ImgFromPackage(fpath))
        logger.info("%s registered as %s", fpath, image_id)

    @classmethod
    def register_from_data(cls, image_id, format, data):
        """Register a image data using image_id"""

        if image_id in cls._stock:
            logger.warning("Warning, replacing resource %s", image_id)
        cls._stock[image_id] = ImgFromData(data, format)
        logger.info("%s registered as %s", "data", image_id)

    @classmethod
    def register_created(cls, image_id, image):
        """Register an already created image using image_id"""

        if image_id in cls._stock:
            logger.warning("Warning, replacing resource {0}", image_id)
        cls._stock[image_id] = ImgCreated(image)
        logger.info("data registered as %s", image_id)

    @classmethod
    def is_registered(cls, image_id):
        return image_id in cls._stock

    @classmethod
    def register_all_from_dir(
        cls, dir_path, prefix=None, ext=None, recurse=False
    ):
        """List files from dir_path and register images with
            filename as key (without extension)

        :param str dir_path: path to search for images.
        :param str prefix: Additionaly a prefix for the key can be provided,
            so the resulting key will be prefix + filename
        :param iterable ext: list of file extensions to load. Defaults to
            tk supported image extensions. Example ('.jpg', '.png')
        :param boolean recurse: search recursivelly.
        """
        prefix = "" if prefix is None else prefix
        if ext is None:
            ext = TK_IMAGE_FORMATS

        path_gen = Path(dir_path).iterdir()
        if recurse:
            path_gen = Path(dir_path).glob("**/*")

        for filename in path_gen:
            if filename.is_file():
                name = filename.stem
                file_ext = filename.suffix
                if file_ext in ext:
                    fkey = f"{prefix}{name}"
                    cls.register(fkey, filename)

    @classmethod
    def register_all_from_pkg(cls, pkg, prefix=None, ext=None, recurse=False):
        """List files from package and register images with
            filename as key (without extension)

        :param str pkg: package to search for images.
        :param str prefix: Additionaly a prefix for the key can be provided,
            so the resulting key will be prefix + filename
        :param iterable ext: list of file extensions to load. Defaults to
            tk supported image extensions. Example ('.jpg', '.png')
        :param boolean recurse: search recursivelly.
        """
        prefix = "" if prefix is None else prefix
        if ext is None:
            ext = TK_IMAGE_FORMATS
        path_gen = resources.files(pkg).iterdir()
        if recurse:
            path_gen = _iter_package_files(pkg)
        for pkg_path in path_gen:
            if pkg_path.is_file():
                fpath = Path(str(pkg_path))
                name = fpath.stem
                file_ext = fpath.suffix
                if file_ext in ext:
                    image_id = f"{prefix}{name}"
                    cls.register_from_package(image_id, pkg_path)

    @classmethod
    def _load_image(cls, image_id):
        """Load image from file or return the cached instance."""

        cache_info = cls._stock[image_id]
        img = None
        try:
            img = cache_info.create_tkimage()
        except ImageFormatNotSupportedError:
            msg = "Error loading image %s, try installing Pillow module."
            logger.error(msg, image_id)
            img = cls.get("img_not_supported")

        cls._cached[image_id] = img
        logger.info("Loaded resource data for %s.", image_id)
        return img

    @classmethod
    def get(cls, image_id):
        """Get image previously registered with key image_id.
        If key not exist, raise StockImageException
        """

        if image_id in cls._cached:
            logger.info("Resource %s is in cache.", image_id)
            return cls._cached[image_id]
        if image_id in cls._stock:
            img = cls._load_image(image_id)
            return img
        else:
            raise StockImageException(f"StockImage: {image_id} not registered.")

    @classmethod
    def as_iconbitmap(cls, image_id):
        """Get image path for use in iconbitmap property"""
        img = None
        if image_id in cls._stock:
            data = cls._stock[image_id]
            if data["type"] not in ("stock", "data", "image"):
                fpath = data["filename"]
                file_ext = fpath.suffix.lower()

                if file_ext in TK_BITMAP_FORMATS:
                    img = BITMAP_TEMPLATE.format(fpath)
        return img

    @classmethod
    def add_resource_path(cls, path):
        if path not in cls._resource_pit:
            cls._resource_pit[path] = PitType.PATH

    @classmethod
    def add_resource_package(cls, package):
        if package not in cls._resource_pit:
            cls._resource_pit[package] = PitType.PACKAGE

    @classmethod
    def find_and_register(cls, image_id):
        """Find and register image from the resource pits."""
        image_path = None
        pattern = f"*{image_id}"
        for pit, pit_type in cls._resource_pit.items():
            if pit_type == PitType.PATH:
                try:
                    image_path = cls._find_in_path(pit, pattern)
                except TypeError:
                    pass
                if image_path:
                    cls.register(image_id, image_path)
                    break
            else:
                try:
                    image_path = cls._find_in_package(pit, pattern)
                except ModuleNotFoundError:
                    pass
                if image_path:
                    cls.register_from_package(image_id, image_path)
                    break
        if image_path is None:
            msg = f"Error: image {image_id} not found in resource pits."
            raise ImageNotFoundError(msg)

    @staticmethod
    def _find_in_path(path_src, pattern):
        found = None
        for p in Path(path_src).glob("**/*"):
            if p.is_file() and p.match(pattern):
                found = p
                break
        return found

    @staticmethod
    def _find_in_package(pkg_src, pattern):
        found = None
        for r in _iter_package_files(pkg_src):
            p2 = Path(str(r))
            if r.is_file() and p2.match(pattern):
                found = r
                break
        return found



from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from ttkwidgets import ItemsCanvas

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class ItemsCanvasBO(TTKFrame):
    class_ = ItemsCanvas
    container = False
    OPTIONS_CUSTOM = (
        "canvaswidth",
        "canvasheight",
        "callback_add",
        "callback_del",
        "callback_move",
        "function_new",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    command_properties = (
        "callback_add",
        "callback_del",
        "callback_move",
        "function_new",
    )

    def _process_property_value(self, pname, value):
        if pname in ("canvaswidth", "canvasheight"):
            return int(value)
        return super(ItemsCanvasBO, self)._process_property_value(pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname in ("callback_add", "callback_del"):
            args = ("item", "rectangle")
        elif cmd_pname == "callback_move":
            args = ("item", "rectangle", "x", "y")
        elif cmd_pname == "function_new":
            args = ("add_item",)
        else:
            args = super(ItemsCanvasBO, self)._code_define_callback_args(
                cmd_pname, cmd
            )
        return args


_builder_uid = f"{_plugin_uid}.ItemsCanvas"
register_widget(
    _builder_uid,
    ItemsCanvasBO,
    "ItemsCanvas",
    ("ttk", _designer_tab_label),
    group=6,
)

register_custom_property(
    _builder_uid,
    "canvaswidth",
    "integernumber",
    help=_("width of the canvas in pixels"),
)
register_custom_property(
    _builder_uid,
    "canvasheight",
    "integernumber",
    help=_("height of the canvas in pixels"),
)
register_custom_property(
    _builder_uid,
    "callback_add",
    "simplecommandentry",
    help=_(
        "callback for when an item is created, with args: (int item, int rectangle)"
    ),
)
register_custom_property(
    _builder_uid,
    "callback_del",
    "simplecommandentry",
    help=_(
        "callback for when an item is deleted, with args: (int item, int rectangle)"
    ),
)
register_custom_property(
    _builder_uid,
    "callback_move",
    "simplecommandentry",
    help=_(
        "callback for when an item is moved, with args: (int item, int rectangle, int x, int y)"
    ),
)
register_custom_property(
    _builder_uid,
    "function_new",
    "simplecommandentry",
    help=_(
        "user defined function for when an item is created, with arg (add_item).\nWhere add_item is a function of this widget."
    ),
)



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKTreeviewBO, TTKTreeviewColumnBO
from pygubu.plugins.pygubu.scrollbarhelper import TTKSBHelperBO
from ttkwidgets.table import Table

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class TableBO(TTKTreeviewBO):
    class_ = Table
    OPTIONS_SPECIFIC = ("height", "padding", "selectmode")
    OPTIONS_CUSTOM = ("drag_cols", "drag_rows", "sortable")
    properties = (
        TTKTreeviewBO.OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )

    def _process_property_value(self, pname, value):
        final_value = value
        if pname in ("drag_cols", "drag_rows", "sortable"):
            final_value = tk.getboolean(value)
        else:
            final_value = super(TableBO, self)._process_property_value(
                pname, value
            )
        return final_value


_builder_uid = _table_uid = f"{_plugin_uid}.Table"
register_widget(
    _builder_uid, TableBO, "Table", ("ttk", _designer_tab_label), group=6
)

register_custom_property(
    _builder_uid,
    "drag_cols",
    "choice",
    values=("", "true", "false"),
    default_value="true",
    help=_("whether columns are draggable"),
)
register_custom_property(
    _builder_uid,
    "drag_rows",
    "choice",
    values=("", "true", "false"),
    default_value="true",
    help=_("whether rows are draggable"),
)
register_custom_property(
    _builder_uid,
    "sortable",
    "choice",
    values=("", "true", "false"),
    default_value="true",
    help=_(
        "whether columns are sortable by clicking on their headings. The sorting order depends on the type of data (str, float, ...) which can be set with the column method."
    ),
)

TTKSBHelperBO.add_allowed_child(_table_uid)


class TableColumnBO(TTKTreeviewColumnBO):
    OPTIONS_CUSTOM = ("type",)
    properties = (
        TTKTreeviewBO.OPTIONS_STANDARD
        + TTKTreeviewBO.OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    allowed_parents = (_table_uid,)

    def _get_column_properties(self, props):
        colprop = super(TableColumnBO, self)._get_column_properties(props)
        type_value = props.get("type", None)
        if type_value is not None:
            final_value = str
            if type_value in ("str", "int", "float", "bool"):
                final_value = eval(type_value)
            colprop["type"] = final_value
        # Setup required anchor
        if colprop["anchor"] == "":
            colprop["anchor"] = "nw"
        return colprop


_builder_uid = f"{_plugin_uid}.Table.Column"

TableBO.add_allowed_child(_builder_uid)

register_widget(
    _builder_uid,
    TableColumnBO,
    "Table.Column",
    ("ttk", _designer_tab_label),
    group=6,
)

register_custom_property(
    _builder_uid,
    "type",
    "choice",
    values=("", "str", "int", "float", "bool"),
    state="readonly",
)



import importlib
from pygubu.i18n import _
from pygubu.api.v1 import BuilderLoaderPlugin


_designer_tab_label = _("TkinterWeb")
_plugin_uid = "tkinterweb"


class TkinterwebLoader(BuilderLoaderPlugin):
    _module = "pygubu.plugins.tkinterweb.htmlwidgets"

    def do_activate(self) -> bool:
        spec = importlib.util.find_spec("tkinterweb")
        return True if spec is not None else False

    def get_module_for(self, identifier: str) -> str:
        return self._module

    def get_all_modules(self):
        return (self._module,)

    def can_load(self, identifier: str) -> bool:
        return identifier.startswith("tkinterweb.")



# encoding: utf-8
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKToplevel
from pygubu.i18n import _
from pygubu.widgets.dialog import Dialog


class DialogBO(TKToplevel):
    class_ = Dialog
    OPTIONS_STANDARD = TKToplevel.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TKToplevel.OPTIONS_SPECIFIC
    OPTIONS_CUSTOM = TKToplevel.OPTIONS_CUSTOM + ("modal",)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    virtual_events = ("<<DialogClose>>",)

    def layout(self, target=None):
        super(DialogBO, self).layout(self.widget.toplevel)

    def _set_property(self, target_widget, pname, value):
        if pname == "modal":
            modal = False
            value = value.lower()
            if value == "true":
                modal = True
            self.widget.set_modal(modal)
        else:
            super(DialogBO, self)._set_property(
                self.widget.toplevel, pname, value
            )

    def get_child_master(self):
        return self.widget.toplevel

    #
    # Code generation methods
    #
    def code_child_master(self):
        return "{0}.toplevel".format(self.code_identifier())

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "modal":
            code_bag[pname] = '"{0}"'.format(value)
        else:
            super(DialogBO, self)._code_set_property(
                targetid, pname, value, code_bag
            )


_builder_id = "pygubu.builder.widgets.dialog"
register_widget(
    _builder_id, DialogBO, "Dialog", (_("Pygubu Widgets"), "ttk"), group=0
)

_help = _("Determines if dialog is run in normal or modal mode.")
register_custom_property(
    _builder_id,
    "modal",
    "choice",
    values=("true", "false"),
    state="readonly",
    help=_help,
)



from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKScrollbar
import awesometkinter as atk

from ..awesometkinter import _designer_tab_label, _plugin_uid


class SimpleScrollbarBO(TTKScrollbar):
    OPTIONS_STANDARD = tuple(
        set(TTKScrollbar.OPTIONS_STANDARD) - set(("style",))
    )
    OPTIONS_CUSTOM = ("bg", "slider_color", "width")
    properties = (
        OPTIONS_STANDARD + TTKScrollbar.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKScrollbar.ro_properties + OPTIONS_CUSTOM + ("orient",)
    class_ = atk.scrollbar.SimpleScrollbar

    def _process_property_value(self, pname, value):
        if pname in ("width",):
            value_ = 5
            try:
                value_ = int(value)
            except ValueError:
                pass
            return value_
        return super()._process_property_value(pname, value)


_builder_uid = _plugin_uid + ".SimpleScrollbar"
register_widget(
    _builder_uid,
    SimpleScrollbarBO,
    "SimpleScrollbar",
    ("ttk", _designer_tab_label),
)

register_custom_property(_builder_uid, "bg", "colorentry")
register_custom_property(_builder_uid, "slider_color", "colorentry")
register_custom_property(_builder_uid, "width", "naturalnumber")



from pygubu.api.v1 import IPluginBase, IDesignerPlugin
from .preview import CTkToplevelPreviewBO, CTkPreviewBO, CTkFramePreviewBO
from ..ctkbase import _plugin_uid


class CTkDesignerPlugin(IDesignerPlugin):
    def get_preview_builder(self, builder_uid: str):
        if builder_uid == f"{_plugin_uid}.CTkToplevel":
            return CTkToplevelPreviewBO
        if builder_uid == f"{_plugin_uid}.CTk":
            return CTkPreviewBO
        if builder_uid == f"{_plugin_uid}.CTkFrame":
            return CTkFramePreviewBO
        return None

    def get_toplevel_preview_for(
        self, builder_uid: str, widget_id: str, builder, top_master
    ):
        top = None
        toplevel_uids = ("customtkinter.CTkToplevel", "customtkinter.CTk")
        if builder_uid in toplevel_uids:
            top = builder.get_object(widget_id)
        return top

    def configure_for_preview(self, builder_uid: str, widget):
        """Make widget just display with minimal functionality."""

        def _no_op(event=None):
            pass

        if builder_uid.endswith(".CTKEntry"):
            seqlist = ("<FocusOut>", "<FocusIn>")
            for seq in seqlist:
                widget.canvas.bind(seq, _no_op)
        elif builder_uid.endswith(".CTkSlider"):
            seqlist = ("<Enter>", "<Leave>", "<Button-1>", "<B1-Motion>")
            for seq in seqlist:
                widget.canvas.bind(seq, _no_op)
        elif builder_uid.endswith(".CTkOptionMenu"):
            seqlist = ("<Enter>", "<Leave>", "<Button-1>")
            for seq in seqlist:
                widget.canvas.bind(seq, _no_op)
                widget.text_label.bind(seq, _no_op)
        elif builder_uid.endswith(".CTkComboBox"):
            widget.canvas.tag_bind("right_parts", "<Enter>", _no_op)
            widget.canvas.tag_bind("dropdown_arrow", "<Enter>", _no_op)
            widget.canvas.tag_bind("right_parts", "<Leave>", _no_op)
            widget.canvas.tag_bind("dropdown_arrow", "<Leave>", _no_op)
            widget.canvas.tag_bind("right_parts", "<Button-1>", _no_op)
            widget.canvas.tag_bind("dropdown_arrow", "<Button-1>", _no_op)



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKScale
from ttkwidgets.tickscale import TickScale

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class TickScaleBO(TTKScale):
    class_ = TickScale
    OPTIONS_CUSTOM = (
        "digits",
        "labelpos",
        "resolution",
        "showvalue",
        "tickinterval",
        "tickpos",
    )
    properties = (
        TTKScale.OPTIONS_STANDARD + TTKScale.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKScale.ro_properties + ("from_", "to")

    def _process_property_value(self, pname, value):
        final_value = value
        if pname in ("digits",):
            final_value = int(value)
        elif pname in ("from", "to_", "resolution", "tickinterval"):
            final_value = float(value)
        else:
            final_value = super(TickScaleBO, self)._process_property_value(
                pname, value
            )
        return final_value


_builder_uid = f"{_plugin_uid}.TickScale"

register_widget(
    _builder_uid,
    TickScaleBO,
    "TickScale",
    ("ttk", _designer_tab_label),
    group=3,
)

register_custom_property(_builder_uid, "digits", "naturalnumber")
register_custom_property(
    _builder_uid,
    "resolution",
    "realnumber",
    help=_(
        "increment by which the slider can be moved. 0 means continuous sliding."
    ),
)
register_custom_property(
    _builder_uid,
    "tickinterval",
    "realnumber",
    help=_("if not 0, display ticks with the given interval"),
)
register_custom_property(
    _builder_uid,
    "showvalue",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "labelpos",
    "choice",
    help="if showvalue is True, position of the label",
    values=("", tk.N, tk.S, tk.E, tk.W),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "tickpos",
    "choice",
    help=_("if tickinterval is not 0, position of the ticks"),
    values=("", tk.N, tk.S, tk.E, tk.W),
    state="readonly",
)



import logging
from ...component.builderobject import (
    CLASS_MAP,
    WidgetDescription,
    CUSTOM_PROPERTIES,
)

logger = logging.getLogger(__name__)


def register_builder(classname, builder, label=None, tags=None, group=None):
    if label is None:
        label = classname
    if tags is None:
        tags = tuple()
    if group is None:
        group = 9

    CLASS_MAP[classname] = WidgetDescription(
        classname, builder, label, tags, group
    )


def register_widget(classname, builder, label=None, tags=None, group=None):
    return register_builder(classname, builder, label, tags, group)


def register_property(name, description):
    if name in CUSTOM_PROPERTIES:
        CUSTOM_PROPERTIES[name].update(description)
        logger.debug("Updating registered property %s", name)
    else:
        CUSTOM_PROPERTIES[name] = description
        logger.debug("Registered property %s", name)


def register_custom_property(
    builder_uid,
    prop_name,
    editor,
    default_value=None,
    help=None,
    **editor_params,
):
    """Helper function to register a custom property.
    All custom properties are created using internal dynamic editor.
    """
    description = {
        "editor": "dynamic",
        builder_uid: {
            "params": {
                "mode": editor,
            }
        },
    }
    description[builder_uid]["params"].update(editor_params)
    if default_value is not None:
        description[builder_uid]["default"] = default_value
    if help is not None:
        description[builder_uid]["help"] = help
    register_property(prop_name, description)



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKListbox
from ttkwidgets.scrolledlistbox import ScrolledListbox

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class ScrolledListboxBO(TKListbox):
    class_ = ScrolledListbox
    container = False
    OPTIONS_CUSTOM = ("compound", "autohidescrollbar")
    properties = (
        TKListbox.OPTIONS_STANDARD + TKListbox.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TKListbox.ro_properties + OPTIONS_CUSTOM

    def configure(self, target=None):
        super(ScrolledListboxBO, self).configure(self.widget.listbox)

    def _process_property_value(self, pname, value):
        final_value = value
        if pname in ("autohidescrollbar",):
            final_value = tk.getboolean(value)
        elif pname == "compound":
            final_value = tk.RIGHT
            values = {"left": tk.LEFT, "right": tk.RIGHT}
            if value in values:
                final_value = values[value]
        else:
            final_value = super(
                ScrolledListboxBO, self
            )._process_property_value(pname, value)
        return final_value

    def code_configure(self, targetid=None):
        if targetid is None:
            targetid = self.code_identifier()
        newtarget = f"{targetid}.listbox"
        return super(ScrolledListboxBO, self).code_configure(newtarget)


_builder_uid = f"{_plugin_uid}.ScrolledListbox"
register_widget(
    _builder_uid,
    ScrolledListboxBO,
    "ScrolledListbox",
    ("ttk", _designer_tab_label),
    group=3,
)

register_custom_property(
    _builder_uid,
    "compound",
    "choice",
    default_value=tk.RIGHT,
    help="side for the Scrollbar to be on",
    values=("", tk.LEFT, tk.RIGHT),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "autohidescrollbar",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)



from pygubu.api.v1 import BuilderObject
from pygubu.plugins.tk.tkstdwidgets import TKToplevel


class ToplevelPreviewFactory(type):
    def __new__(cls, clsname, superclasses, attrs):
        return type.__new__(cls, str(clsname), superclasses, attrs)


class ToplevelPreviewMixin(object):
    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)
        self.tl_attrs = {}
        self._w_set = False
        self._h_set = False

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "width"
        if key in kw:
            value = int(kw[key])
            minsize = self.tl_attrs.get("minsize", None)
            maxsize = self.tl_attrs.get("maxsize", None)
            remove = False
            if minsize and value < minsize[0]:
                remove = True
            if maxsize and value > maxsize[0]:
                remove = True
            if self._w_set:
                resizable = self.tl_attrs.get("resizable", None)
                if resizable and not TKToplevel.RESIZABLE[resizable][0]:
                    remove = True
            if remove:
                kw.pop(key)
            else:
                self._w_set = True
        key = "height"
        if key in kw:
            value = int(kw[key])
            minsize = self.tl_attrs.get("minsize", None)
            maxsize = self.tl_attrs.get("maxsize", None)
            remove = False
            if minsize and value < minsize[1]:
                remove = True
            if maxsize and value > maxsize[1]:
                remove = True
            if self._h_set:
                resizable = self.tl_attrs.get("resizable", None)
                if resizable and not TKToplevel.RESIZABLE[resizable][1]:
                    remove = True
            if remove:
                kw.pop(key)
            else:
                self._h_set = True
        key = "menu"
        if key in kw:
            # No menu preview available
            kw.pop(key)
        return super().configure(cnf, **kw)


class ToplevelPreviewBaseBO(BuilderObject):
    container = True
    container_layout = True
    # Add fake 'modal' property for Dialog preview
    properties = TKToplevel.properties + ("modal",)
    ro_properties = TKToplevel.ro_properties

    def configure(self, target=None):
        # setup width and height properties if
        # geometry is defined.
        geom = "geometry"
        if geom in self.wmeta.properties:
            w, h = self._get_dimwh(self.wmeta.properties[geom])
            if w and h:
                self.wmeta.properties["width"] = w
                self.wmeta.properties["height"] = h
        super().configure(target)

    def configure_children(self, target=None):
        w = self.widget
        # Set a fixed size in preview if geometry is set.
        if "geometry" in self.wmeta.properties:
            if w.pack_slaves():
                w.pack_propagate(0)
            elif w.grid_slaves():
                w.grid_propagate(0)

    def layout(self, target=None):
        self.widget.pack(expand=True, fill="both")
        self._container_layout(
            self.widget,
            self.wmeta.container_manager,
            self.wmeta.container_properties,
        )

    def _get_dimwh(self, dimvalue: str):
        # get width and height from dimension string
        dim = dimvalue.split("+")[0]
        dim = dim.split("-")[0]
        w, h = dim.split("x")
        return (w, h)

    def _set_property(self, target_widget, pname, value):
        tw = target_widget
        tw.tl_attrs[pname] = value
        method_props = (
            "iconbitmap",
            "iconphoto",
            "overrideredirect",
            "title",
        )
        if pname in method_props:
            pass
        elif pname in ("maxsize", "minsize"):
            if not value:
                del tw.tl_attrs[pname]
            elif "|" in value:
                w, h = value.split("|")
                if w and h:
                    tw.tl_attrs[pname] = (int(w), int(h))
                else:
                    del tw.tl_attrs[pname]
        elif pname == "geometry":
            if value:
                w, h = self._get_dimwh(value)
                if w and h:
                    tw.tl_attrs["minsize"] = (int(w), int(h))
                    tw._h_set = tw._w_set = False
                    tw.configure(width=w, height=h)
        elif pname == "resizable":
            # Do nothing, fake 'resizable' property for Toplevel preview
            pass
        elif pname == "modal":
            # Do nothing, fake 'modal' property for dialog preview
            pass
        else:
            super()._set_property(tw, pname, value)



""" Widget utility functions."""


def crop_widget(widget, *, recursive=False):
    """Remove standard widget functionality."""
    wclass = widget.winfo_class()
    bindtags = widget.bindtags()
    if wclass in bindtags:
        bindtags = list(bindtags)
        bindtags.remove(wclass)
        widget.bindtags(bindtags)
    if recursive:
        for childw in widget.winfo_children():
            crop_widget(childw, recursive=recursive)



from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.i18n import _
from pygubu.plugins.tk.tkstdwidgets import TKToplevel as TKToplevelBO
from customtkinter import (
    CTk,
    CTkToplevel,
    set_appearance_mode,
    set_default_color_theme,
)

from ..customtkinter import _designer_tab_label, _plugin_uid
from .ctkbase import CTkBaseMixin


class CTkBO(TKToplevelBO):
    class_ = CTk
    properties = TKToplevelBO.properties + (
        "appearance_mode",
        "color_theme",
        "fg_color",
    )
    ro_properties = TKToplevelBO.ro_properties + ("background", "fg_color")

    def realize(self, parent):
        # Call realize from BuilderObject not Toplevel.
        return super(TKToplevelBO, self).realize(parent)

    def _set_property(self, target_widget, pname, value):
        if pname == "appearance_mode":
            set_appearance_mode(value)
        elif pname == "color_theme":
            set_default_color_theme(value)
        else:
            return super()._set_property(target_widget, pname, value)

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "appearance_mode":
            code_bag[pname] = (f'set_appearance_mode("{value}")',)
        elif pname == "color_theme":
            code_bag[pname] = (f'set_default_color_theme("{value}")',)
        else:
            super()._code_set_property(targetid, pname, value, code_bag)

    def code_imports(self):
        imports = ["CTk"]
        if "appearance_mode" in self.wmeta.properties:
            imports.append("set_appearance_mode")
        if "color_theme" in self.wmeta.properties:
            imports.append("set_default_color_theme")
        return [("customtkinter", name) for name in imports]


_builder_uid = f"{_plugin_uid}.CTk"
register_widget(
    _builder_uid,
    CTkBO,
    "CTk",
    ("ttk", _designer_tab_label),
    group=-1,
)

_maxsize_help = _("Set the maximum window size.")
_minsize_help = _("Set the minimum window size.")

register_custom_property(
    _builder_uid,
    "minsize",
    "whentry",
    help=_minsize_help,
)
register_custom_property(
    _builder_uid,
    "maxsize",
    "whentry",
    help=_maxsize_help,
)


class CTkToplevelBO(TKToplevelBO):
    class_ = CTkToplevel
    properties = TKToplevelBO.properties + ("fg_color",)
    ro_properties = TKToplevelBO.ro_properties + ("background", "fg_color")

    def realize(self, parent):
        # Call realize from BuilderObject not Toplevel.
        return super(TKToplevelBO, self).realize(parent)

    def code_imports(self):
        return [("customtkinter", self.class_.__name__)]


_builder_uid = f"{_plugin_uid}.CTkToplevel"
register_widget(
    _builder_uid,
    CTkToplevelBO,
    "CTkToplevel",
    ("ttk", _designer_tab_label),
    group=-1,
)

register_custom_property(
    _builder_uid,
    "minsize",
    "whentry",
    help=_minsize_help,
)
register_custom_property(
    _builder_uid,
    "maxsize",
    "whentry",
    help=_maxsize_help,
)



# encoding: utf-8
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from pygubu.widgets.pathchooserinput import PathChooserInput


class PathChooserInputBuilder(BuilderObject):
    class_ = PathChooserInput
    OPTIONS_CUSTOM = (
        "type",
        "path",
        "image",
        "textvariable",
        "state",
        "initialdir",
        "mustexist",
        "title",
    )
    properties = OPTIONS_CUSTOM
    virtual_events = ("<<PathChooserPathChanged>>",)

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "type":
            code_bag[pname] = '"{0}"'.format(value)
        elif pname in ("initialdir", "mustexist", "title"):
            code_bag[pname] = '"{0}"'.format(value)
        elif pname == "textvariable":
            code_bag[pname] = self._code_set_tkvariable_property(pname, value)
        else:
            super(PathChooserInputBuilder, self)._code_set_property(
                targetid, pname, value, code_bag
            )


_builder_id = "pygubu.builder.widgets.pathchooserinput"
register_widget(
    _builder_id,
    PathChooserInputBuilder,
    "PathChooserInput",
    ("ttk", _("Pygubu Widgets")),
)

_help = "Dialog type"
register_custom_property(
    _builder_id,
    "type",
    "choice",
    values=(PathChooserInput.FILE, PathChooserInput.DIR),
    state="readonly",
    default_value=PathChooserInput.FILE,
    help=_help,
)

_help = "Initial path value."
register_custom_property(_builder_id, "path", "entry", help=_help)

_help = "Image for the button."
register_custom_property(_builder_id, "image", "imageentry", help=_help)

_help = "Tk variable associated to the path property."
register_custom_property(_builder_id, "textvariable", "tkvarentry", help=_help)

_help = "Path entry state."
register_custom_property(
    _builder_id,
    "state",
    "choice",
    values=("", "normal", "disabled", "readonly"),
    state="readonly",
    help=_help,
)

_help = "Dialog option. Determines if path must exist for directory dialog."
register_custom_property(
    _builder_id,
    "mustexist",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_help,
)

_help = "Dialog option. Sets initial directory."
register_custom_property(_builder_id, "initialdir", "entry", help=_help)

_help = "Dialog option. Sets dialog title."
register_custom_property(_builder_id, "title", "entry", help=_help)



from pygubu.api.v1 import BuilderLoaderPlugin


class PygubuWidgetsLoader(BuilderLoaderPlugin):
    builders = {
        "pygubu.builder.widgets.calendarframe": "pygubu.plugins.pygubu.calendarframe",
        "pygubu.builder.widgets.combobox": "pygubu.plugins.pygubu.combobox",
        "pygubu.builder.widgets.dialog": "pygubu.plugins.pygubu.dialog",
        "pygubu.builder.widgets.editabletreeview": "pygubu.plugins.pygubu.editabletreeview",
        "pygubu.builder.widgets.pathchooserinput": "pygubu.plugins.pygubu.pathchooserinput",
        "pygubu.builder.widgets.scrollbarhelper": "pygubu.plugins.pygubu.scrollbarhelper",
        "pygubu.builder.widgets.scrolledframe": "pygubu.plugins.pygubu.scrolledframe",
        "pygubu.builder.widgets.tkinterscrolledtext": "pygubu.plugins.pygubu.tkinterscrolledtext",
        "pygubu.builder.widgets.tkscrollbarhelper": "pygubu.plugins.pygubu.tkscrollbarhelper",
        "pygubu.builder.widgets.tkscrolledframe": "pygubu.plugins.pygubu.tkscrolledframe",
    }

    def do_activate(self) -> bool:
        return True

    def get_all_modules(self):
        return [m for m in self.builders.values()]

    def get_module_for(self, identifier: str) -> str:
        return self.builders[identifier]

    def can_load(self, identifier: str) -> bool:
        return identifier in self.builders

    def get_designer_plugin(self):
        """Load class that implements IDesignerPlugin"""
        from .designer.designerplugin import PygubuDesignerPlugin

        return PygubuDesignerPlugin()



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKListbox
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame, TTKCombobox
from ttkwidgets.font import (
    FontFamilyDropdown,
    FontFamilyListbox,
    FontSelectFrame,
    FontPropertiesFrame,
    FontSizeDropdown,
)

from ..ttkwidgets import _designer_tab_label, _plugin_uid
from .utils import CallbakInitArgMixin
from .autocomplete import AutocompleteComboboxBO
from .scrolledlistbox import ScrolledListboxBO


class FontFamilyDropdownBO(CallbakInitArgMixin, AutocompleteComboboxBO):
    class_ = FontFamilyDropdown
    init_completevalues = False
    container = False
    OPTIONS_SPECIFIC = tuple(
        set(AutocompleteComboboxBO.OPTIONS_SPECIFIC) - set(("textvariable",))
    )
    OPTIONS_CUSTOM = ("callback",)
    properties = (
        AutocompleteComboboxBO.OPTIONS_STANDARD
        + OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    ro_properties = TTKCombobox.ro_properties
    command_properties = ("callback",)

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname == "callback":
            args = ("family",)
        else:
            args = super()._code_define_callback_args(cmd_pname, cmd)
        return args


_builder_uid = f"{_plugin_uid}.FontFamilyDropdown"
register_widget(
    _builder_uid,
    FontFamilyDropdownBO,
    "FontFamilyDropdown",
    ("ttk", _designer_tab_label),
)

register_custom_property(
    _builder_uid,
    "callback",
    "simplecommandentry",
    help=_(
        "name of the callback function with one argument: the font family name"
    ),
)


class FontFamilyListboxBO(CallbakInitArgMixin, ScrolledListboxBO):
    class_ = FontFamilyListbox
    container = False
    OPTIONS_STANDARD = tuple(
        set(TKListbox.OPTIONS_STANDARD)
        - set(("xscrollcommand", "yscrollcommand"))
    )
    OPTIONS_SPECIFIC = tuple(
        set(TKListbox.OPTIONS_SPECIFIC) - set(("listvariable",))
    )
    OPTIONS_CUSTOM = ("autohidescrollbar", "callback")
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = TKListbox.ro_properties + ("autohidescrollbar",)
    command_properties = ("callback",)

    def _process_property_value(self, pname, value):
        if pname == "autohidescrollbar":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value):
        if pname == "autohidescrollbar":
            return tk.getboolean(value)
        return super()._code_process_property_value(targetid, pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname == "callback":
            args = ("family",)
        else:
            args = super()._code_define_callback_args(cmd_pname, cmd)
        return args


_builder_uid = f"{_plugin_uid}.FontFamilyListbox"
register_widget(
    _builder_uid,
    FontFamilyListboxBO,
    "FontFamilyListbox",
    ("ttk", _designer_tab_label),
)

register_custom_property(
    _builder_uid,
    "autohidescrollbar",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "callback",
    "simplecommandentry",
    help=_(
        "name of the callback function with one argument: the font family name"
    ),
)


class FontSelectFrameBO(CallbakInitArgMixin, TTKFrame):
    class_ = FontSelectFrame
    container = False
    OPTIONS_CUSTOM = ("callback",)
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKFrame.ro_properties + OPTIONS_CUSTOM
    command_properties = ("callback",)

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname == "callback":
            args = ("family", "size", "bold", "italic", "underline")
        else:
            args = super(FontSelectFrameBO, self)._code_define_callback_args(
                cmd_pname, cmd
            )
        return args


_builder_uid = f"{_plugin_uid}.FontSelectFrame"
register_widget(
    _builder_uid,
    FontSelectFrameBO,
    "FontSelectFrame",
    ("ttk", _designer_tab_label),
)

register_custom_property(
    _builder_uid,
    "callback",
    "simplecommandentry",
    help=_(
        "name of the callback function with arguments: (family: str, size: int, bold: bool, italic: bool, underline: bool)"
    ),
)


class FontPropertiesFrameBO(CallbakInitArgMixin, TTKFrame):
    class_ = FontPropertiesFrame
    container = False
    OPTIONS_CUSTOM = ("callback", "label", "fontsize")
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKFrame.ro_properties + OPTIONS_CUSTOM
    command_properties = ("callback",)

    def _process_property_value(self, pname, value):
        if pname == "label":
            return tk.getboolean(value)
        if pname == "fontsize":
            return int(value)
        return super()._process_property_value(pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        if cmd_pname == "callback":
            return ("family", "size", "bold", "italic", "underline")
        return super()._code_define_callback_args(cmd_pname, cmd)


_builder_uid = f"{_plugin_uid}.FontPropertiesFrame"
register_widget(
    _builder_uid,
    FontPropertiesFrameBO,
    "FontPropertiesFrame",
    ("ttk", _designer_tab_label),
)

register_custom_property(
    _builder_uid,
    "callback",
    "simplecommandentry",
    help=_(
        "name of the callback function with arguments: (bold: bool, italic: bool, underline: bool, overstrike: bool)"
    ),
)
register_custom_property(
    _builder_uid,
    "label",
    "choice",
    values=("", "true", "false"),
    default_value="true",
    state="readonly",
    help=_("Show/hide Header label"),
)
register_custom_property(
    _builder_uid,
    "fontsize",
    "naturalnumber",
    help=_("size of the font on the buttons"),
)


class FontSizeDropdownBO(CallbakInitArgMixin, AutocompleteComboboxBO):
    class_ = FontSizeDropdown
    init_completevalues = False
    container = False
    OPTIONS_SPECIFIC = tuple(
        set(AutocompleteComboboxBO.OPTIONS_SPECIFIC) - set(("textvariable",))
    )
    OPTIONS_CUSTOM = ("callback",)
    properties = (
        AutocompleteComboboxBO.OPTIONS_STANDARD
        + OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    ro_properties = TTKCombobox.ro_properties
    command_properties = ("callback",)

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname == "callback":
            args = ("size",)
        else:
            args = super(FontSizeDropdownBO, self)._code_define_callback_args(
                cmd_pname, cmd
            )
        return args


_builder_uid = f"{_plugin_uid}.FontSizeDropdown"
register_widget(
    _builder_uid,
    FontSizeDropdownBO,
    "FontSizeDropdown",
    ("ttk", _designer_tab_label),
)

register_custom_property(
    _builder_uid,
    "callback",
    "simplecommandentry",
    help=_(
        "name of the callback function on click with single argument: size: int"
    ),
)



# encoding: utf-8
import tkinter as tk

from pygubu.api.v1 import BuilderObject, register_widget
from pygubu.i18n import _
from pygubu.widgets.tkscrollbarhelper import TkScrollbarHelper


class TKSBHelperBO(BuilderObject):
    class_ = TkScrollbarHelper
    container = True
    maxchildren = 1
    allowed_children = (
        "tk.Entry",
        "ttk.Entry",
        "tk.Text",
        "tk.Canvas",
        "tk.Listbox",
        "ttk.Treeview",
    )
    OPTIONS_STANDARD = (
        "borderwidth",
        "cursor",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "padx",
        "pady",
        "relief",
        "takefocus",
    )
    OPTIONS_SPECIFIC = (
        "background",
        "class_",
        "container",
        "height",
        "width",
    )
    OPTIONS_CUSTOM = ("scrolltype", "usemousewheel")
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = ("class_", "scrolltype")
    allow_bindings = False

    def get_child_master(self):
        return self.widget.container

    def add_child(self, bobject):
        cwidget = bobject.widget
        self.widget.add_child(cwidget)

    #
    # Code generation methods
    #
    def code_child_master(self):
        return "{0}.container".format(self.code_identifier())

    def code_child_add(self, childid):
        lines = []
        line = "{0}.add_child({1})".format(self.code_identifier(), childid)
        lines.append(line)
        return lines

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "usemousewheel":
            code_bag[pname] = "{0}".format(tk.getboolean(value))
        else:
            super(TKSBHelperBO, self)._code_set_property(
                targetid, pname, value, code_bag
            )


register_widget(
    "pygubu.builder.widgets.tkscrollbarhelper",
    TKSBHelperBO,
    "ScrollbarHelper",
    (_("Pygubu Helpers"), "tk"),
)



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKText
import awesometkinter as atk

from ..awesometkinter import _designer_tab_label, _plugin_uid


class ScrolledTextBO(TKText):
    OPTIONS_CUSTOM = TKText.OPTIONS_CUSTOM + (
        "bg",
        "fg",
        "bd",
        "vscroll",
        "hscroll",
        "autoscroll",
        "max_chars",
        "sbar_fg",
        "sbar_bg",
        "vbar_width",
        "hbar_width",
    )
    properties = (
        TKText.OPTIONS_STANDARD + TKText.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = tuple(
        (set(TKText.ro_properties) | set(OPTIONS_CUSTOM) | set(("wrap",)))
        - set(("text",))
    )
    class_ = atk.text.ScrolledText

    def _process_property_value(self, pname, value):
        if pname in ("vscroll", "hscroll", "autoscroll"):
            return tk.getboolean(value)
        if pname in ("max_chars", "vbar_width", "hbar_width"):
            return int(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value):
        if pname in ("vscroll", "hscroll", "autoscroll"):
            return tk.getboolean(value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = _plugin_uid + ".ScrolledText"
register_widget(
    _builder_uid,
    ScrolledTextBO,
    "ScrolledText",
    ("ttk", _designer_tab_label),
    group=1,
)
register_custom_property(
    _builder_uid, "bg", "colorentry", help=_("background color")
)
register_custom_property(
    _builder_uid, "fg", "colorentry", help=_("foreground color")
)
register_custom_property(
    _builder_uid, "bd", "naturalnumber", help=_("border width")
)
register_custom_property(
    _builder_uid,
    "vscroll",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_("include vertical scrollbar"),
)
register_custom_property(
    _builder_uid,
    "hscroll",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_("include horizontal scrollbar"),
)
register_custom_property(
    _builder_uid,
    "autoscroll",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_("automatic vertical scrolling"),
)
register_custom_property(
    _builder_uid,
    "max_chars",
    "naturalnumber",
    help=_(
        "maximum characters allowed in Text widget, text will be truncated from the beginning to match the max chars"
    ),
)
register_custom_property(
    _builder_uid, "sbar_fg", "colorentry", help=_("color of scrollbars' slider")
)
register_custom_property(
    _builder_uid,
    "sbar_bg",
    "colorentry",
    help=_("color of scrollbars' trough, default to frame's background"),
)
register_custom_property(
    _builder_uid,
    "vbar_width",
    "naturalnumber",
    help=_("vertical scrollbar width"),
)
register_custom_property(
    _builder_uid,
    "hbar_width",
    "naturalnumber",
    help=_("horizontal scrollbar width"),
)



# encoding: utf-8
import json
import logging
import tkinter as tk
from collections import defaultdict, namedtuple

__all__ = [
    "BuilderObject",
    "EntryBaseBO",
    "PanedWindowBO",
    "PanedWindowPaneBO",
    "WidgetDescription",
    "CLASS_MAP",
    "CB_TYPES",
    "CUSTOM_PROPERTIES",
    "TRANSLATABLE_PROPERTIES",
]

logger = logging.getLogger(__name__)

#
# Utility functions
#


def isfloat(num: str) -> bool:
    try:
        float(num)
        return True
    except ValueError:
        return False


#
# BuilderObject
#

WidgetDescription = namedtuple(
    "WidgetDescription", ["classname", "builder", "label", "tags", "group"]
)

CLASS_MAP = {}
CUSTOM_PROPERTIES = {}


class CB_TYPES:
    """Callback types"""

    SIMPLE = "simple"
    WITH_WID = "with_wid"
    ENTRY_VALIDATE = "entry_validate"
    SCROLL = "scroll"
    SCROLLSET = "scrollset"
    SCALE = "scale"
    BIND_EVENT = "bind_event"


TRANSLATABLE_PROPERTIES = ["label", "text", "title"]

#
# Base class
#


class BuilderObject(object):
    """Base class for Widgets created with Builder"""

    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = tuple()
    OPTIONS_CUSTOM = tuple()
    class_ = None
    container = False
    container_layout = False
    allowed_parents = None
    allowed_children = None
    maxchildren = None
    properties = tuple()
    ro_properties = tuple()
    layout_required = True
    command_properties = tuple()
    allow_bindings = True
    tkvar_properties = ("listvariable", "textvariable", "variable")
    tkimage_properties = (
        "image",
        "selectimage",
        "iconphoto",
        "backgroundimage",
    )
    tkfont_properties = ("font",)
    virtual_events = tuple()

    @classmethod
    def factory(cls, builder, wdata):
        clsobj = cls(builder, wdata)
        wdata.layout_required = clsobj.layout_required
        return clsobj

    @classmethod
    def add_allowed_parent(cls, builder_uid):
        if cls.allowed_parents is None:
            cls.allowed_parents = (builder_uid,)
        else:
            cls.allowed_parents = cls.allowed_parents + (builder_uid,)

    @classmethod
    def add_allowed_child(cls, builder_uid):
        if cls.allowed_children is None:
            cls.allowed_children = (builder_uid,)
        else:
            cls.allowed_children = cls.allowed_children + (builder_uid,)

    def __init__(self, builder, wmeta):
        super(BuilderObject, self).__init__()
        self.widget = None
        self.builder = builder
        self.wmeta = wmeta
        self._code_identifier = None

    def realize(self, parent):
        args = self._get_init_args()
        master = parent.get_child_master()
        self.widget = self.class_(master, **args)
        return self.widget

    def _get_init_args(self):
        """Creates dict with properties marked as readonly"""

        args = {}
        for rop in self.ro_properties:
            if rop in self.wmeta.properties:
                pvalue = self._process_property_value(
                    rop, self.wmeta.properties[rop]
                )
                args[rop] = pvalue
        return args

    def configure(self, target=None):
        if target is None:
            target = self.widget
        for pname, value in self.wmeta.properties.items():
            if (
                pname not in self.ro_properties
                and pname not in self.command_properties
            ):
                self._set_property(target, pname, value)

    def configure_children(self, target=None):
        "Aditional configurations after adding all children."
        pass

    def _process_property_value(self, pname, value):
        propvalue = value
        if pname in self.tkvar_properties:
            propvalue = self.builder.create_variable(value)
            if "text" in self.wmeta.properties and pname == "textvariable":
                propvalue.set(self.wmeta.properties["text"])
            elif "value" in self.wmeta.properties and pname == "variable":
                propvalue.set(self.wmeta.properties["value"])
        elif pname in self.tkimage_properties:
            propvalue = self.builder.get_image(value)
        elif pname == "takefocus":
            if value:
                propvalue = tk.getboolean(value)
        return propvalue

    def _set_property(self, target_widget, pname, value):
        if pname not in self.__class__.properties:
            msg = "Attempt to set an unknown property '%s' on class '%s'"
            logger.warning(msg, pname, repr(self.class_))
        else:
            try:
                logger.debug("processing property %s value", pname)
                propvalue = self._process_property_value(pname, value)
                logger.debug("setting property %s = %s", pname, propvalue)
                target_widget.configure(**{pname: propvalue})
            except Exception as e:
                msg = "Failed to set property '%s' on class '%s'. Error: %s"
                logger.error(msg, pname, repr(self.class_), str(e))
                # logger.exception(e)

    def layout(self, target=None):
        if target is None:
            target = self.widget
        if self.layout_required:
            # Check manager
            manager = self.wmeta.manager
            logger.debug(
                "Applying %s layout to %s", manager, self.wmeta.identifier
            )
            properties = self.wmeta.layout_properties
            if manager == "grid":
                target.grid(**properties)
            elif manager == "pack":
                target.pack(**properties)
            elif manager == "place":
                target.place(**self.wmeta.layout_properties)
            else:
                msg = f"Invalid layout manager: {manager}"
                raise Exception(msg)

        # container layout
        if self.container_layout:
            self._container_layout(
                target,
                self.wmeta.container_manager,
                self.wmeta.container_properties,
            )

    def _container_layout(self, target, container_manager, properties):
        # Do container layout
        propagate = properties.get("propagate", "true")
        propagate = tk.getboolean(propagate)
        anchor = properties.get("anchor", None)

        if container_manager == "grid":
            if anchor:
                target.grid_anchor(anchor)
            if not propagate:
                target.grid_propagate(0)
            self._gridrc_config(target)
        elif container_manager == "pack":
            if not propagate:
                target.pack_propagate(0)
        elif container_manager is None:
            raise Exception("Container Manager is none :(")

    def _gridrc_config(self, target):
        # configure grid row/col properties:
        for type_, num, pname, value in self.wmeta.gridrc_properties:
            if type_ == "row":
                target.rowconfigure(num, **{pname: value})
            else:
                target.columnconfigure(num, **{pname: value})

    def get_child_master(self):
        return self.widget

    def add_child(self, bobject):
        pass

    def _create_callback(self, cmd, command):
        callback = command
        cmd_type = cmd["cbtype"]
        if cmd_type == CB_TYPES.WITH_WID:

            def widget_callback(button_id=self.wmeta.identifier):
                command(button_id)

            callback = widget_callback
        if cmd_type == CB_TYPES.ENTRY_VALIDATE:
            args = cmd["args"]
            if args:
                args = args.split(" ")
                callback = (self.widget.register(command),) + tuple(args)
            else:
                callback = self.widget.register(command)
        return callback

    def _connect_command(self, cmd_pname, callback):
        prop = {cmd_pname: callback}
        self.widget.configure(**prop)

    def connect_commands(self, cmd_bag):
        notconnected = []
        commands = {}
        for cmd_pname in self.command_properties:
            cmd = self.wmeta.properties.get(cmd_pname, None)
            if cmd is not None:
                cmd = json.loads(cmd)
                cmd_name = (cmd["value"]).strip()
                if cmd_name:
                    commands[cmd_pname] = cmd
                else:
                    msg = "%s: invalid value for property '%s'."
                    logger.warning(msg, self.wmeta.identifier, cmd_pname)

        if isinstance(cmd_bag, dict):
            for cmd_pname, cmd in commands.items():
                cmd_name = cmd["value"]
                if cmd_name in cmd_bag:
                    callback = self._create_callback(cmd, cmd_bag[cmd_name])
                    self._connect_command(cmd_pname, callback)
                else:
                    notconnected.append(cmd_name)
        else:
            for cmd_pname, cmd in commands.items():
                cmd_name = cmd["value"]
                if hasattr(cmd_bag, cmd_name):
                    callback = self._create_callback(
                        cmd, getattr(cmd_bag, cmd_name)
                    )
                    self._connect_command(cmd_pname, callback)
                else:
                    notconnected.append(cmd_name)
        if notconnected:
            return notconnected
        else:
            return None

    def connect_bindings(self, cb_bag):
        notconnected = []

        if isinstance(cb_bag, dict):
            for bind in self.wmeta.bindings:
                cb_name = bind.handler
                if cb_name in cb_bag:
                    callback = cb_bag[cb_name]
                    self.widget.bind(bind.sequence, callback, add=bind.add)
                else:
                    notconnected.append(cb_name)
        else:
            for bind in self.wmeta.bindings:
                cb_name = bind.handler
                if hasattr(cb_bag, cb_name):
                    callback = getattr(cb_bag, cb_name)
                    self.widget.bind(bind.sequence, callback, add=bind.add)
                else:
                    notconnected.append(cb_name)
        if notconnected:
            return notconnected
        else:
            return None

    #
    # Code generation methods
    #
    @staticmethod
    def code_escape_str(value: str) -> str:
        return repr(value)

    def _code_get_init_args(self, code_identifier):
        """Creates dict with properties marked as readonly"""
        args = {}
        for rop in self.ro_properties:
            if rop in self.wmeta.properties:
                pvalue = self._code_process_property_value(
                    code_identifier, rop, self.wmeta.properties[rop]
                )
                args[rop] = pvalue
        return args

    def code_realize(self, boparent, code_identifier=None):
        if code_identifier is not None:
            self._code_identifier = code_identifier
        lines = []
        master = boparent.code_child_master()
        init_args = self._code_get_init_args(self.code_identifier())
        bag = []
        for pname, value in init_args.items():
            bag.append(f"{pname}={value}")
        kwargs = ""
        if bag:
            kwargs = f""", {", ".join(bag)}"""
        s = f"{self.code_identifier()} = {self._code_class_name()}({master}{kwargs})"
        lines.append(s)
        return lines

    def code_identifier(self):
        if self._code_identifier is None:
            return self.wmeta.identifier
        return self._code_identifier

    def code_child_master(self):
        return self.code_identifier()

    def code_child_add(self, childid):
        return []

    def code_configure(self, targetid=None):
        if targetid is None:
            targetid = self.code_identifier()
        (
            code_bag,
            kwproperties,
            complex_properties,
        ) = self._code_process_properties(self.wmeta.properties, targetid)
        lines = []
        args_bag = []
        for pname in sorted(kwproperties):
            arg_stmt = f"{pname}={code_bag[pname]}"
            args_bag.append(arg_stmt)
        if args_bag:
            args = ", ".join(args_bag)
            line = f"{targetid}.configure({args})"
            lines.append(line)
        for pname in complex_properties:
            lines.extend(code_bag[pname])

        return lines

    def code_configure_children(self, target=None):
        "Aditional configurations after adding all children."
        return tuple()

    def code_layout(self, targetid=None, parentid=None):
        if targetid is None:
            targetid = self.code_identifier()
        if parentid is None:
            parentid = targetid
        lines = []
        if self.layout_required:
            manager = self.wmeta.manager
            layout = self.wmeta.layout_properties
            args_bag = []
            for p, v in sorted(layout.items()):
                pvalue = self._code_process_layout_property(manager, p, v)
                arg_stmt = f"{p}={pvalue}"
                args_bag.append(arg_stmt)
            args = ", ".join(args_bag)

            line = f"{targetid}.{manager}({args})"
            lines.append(line)

        if self.container_layout:
            manager = self.wmeta.container_manager
            container_prop = self.wmeta.container_properties
            pvalue = str(container_prop.get("propagate", "")).lower()
            if "propagate" in container_prop and pvalue == "false":
                line = f"{targetid}.{manager}_propagate(0)"
                lines.append(line)
            if manager == "grid" and "anchor" in container_prop:
                value = container_prop.get("anchor")
                line = f'{targetid}.{manager}_anchor("{value}")'
                lines.append(line)

            rowbag = defaultdict(list)
            colbag = defaultdict(list)
            for type_, num, pname, value in self.wmeta.gridrc_properties:
                pvalue = self._code_process_layout_property(
                    "grid", pname, value
                )
                arg = f"{pname}={pvalue}"
                if type_ == "row":
                    rowbag[num].append(arg)
                else:
                    colbag[num].append(arg)
            for k, bag in rowbag.items():
                args = ", ".join(bag)
                rowid = f'"{k}"' if k == "all" else k
                line = f"{targetid}.rowconfigure({rowid}, {args})"
                lines.append(line)
            for k, bag in colbag.items():
                args = ", ".join(bag)
                colid = f'"{k}"' if k == "all" else k
                line = f"{targetid}.columnconfigure({colid}, {args})"
                lines.append(line)
        return lines

    def _code_class_name(self):
        cname = None
        # try here ?
        cname = self.builder.code_classname_for(self)
        if cname is None:
            if self.class_ is not None:
                prefix = self.class_.__module__
                name = self.class_.__name__
                cname = f"{prefix}.{name}"
            else:
                cname = "ClassNameNotDefined"
        return cname

    def _code_process_properties(
        self, properties, targetid, skip_ro=True, skip_commands=True
    ):
        code_bag = {}
        for pname, value in properties.items():
            if pname in self.ro_properties and skip_ro:
                continue
            if pname in self.command_properties and skip_commands:
                continue
            self._code_set_property(targetid, pname, value, code_bag)

        # properties
        # determine kw properties or complex properties
        kwproperties = []
        complex_properties = []
        for pname, value in code_bag.items():
            if (
                isinstance(value, str)
                or isinstance(value, bool)
                or isinstance(value, int)
                or isinstance(value, float)
            ):
                kwproperties.append(pname)
            else:
                complex_properties.append(pname)

        return (code_bag, kwproperties, complex_properties)

    def _code_process_property_value(self, targetid, pname, value: str):
        propvalue = None
        if pname in self.tkvar_properties:
            propvalue = self._code_set_tkvariable_property(pname, value)
        elif pname in self.command_properties:
            cmd = json.loads(value)
            propvalue = self._code_define_callback(pname, cmd)
        elif pname in self.tkimage_properties:
            propvalue = self.builder.code_create_image(value)
        elif pname == "takefocus":
            propvalue = str(tk.getboolean(value))
        elif pname in TRANSLATABLE_PROPERTIES:
            propvalue = self.builder.code_translate_str(value)
        # default processing
        if propvalue is None:
            pvalue_str = str(value)
            propvalue = (
                pvalue_str
                if pvalue_str.isnumeric() or isfloat(pvalue_str)
                else f'"{value}"'
            )
        return propvalue

    def _code_set_property(self, targetid, pname, value, code_bag):
        code_bag[pname] = self._code_process_property_value(
            targetid, pname, value
        )

    def _code_set_tkvariable_property(self, pname, value):
        """Create code for tk variable property.
        Can be used from subclases for custom tk variable properties."""
        varvalue = None
        if "text" in self.wmeta.properties and pname == "textvariable":
            varvalue = self.wmeta.properties["text"]
        elif "value" in self.wmeta.properties and pname == "variable":
            varvalue = self.wmeta.properties["value"]
        propvalue = self.builder.code_create_variable(value, varvalue)
        return propvalue

    def code_connect_commands(self):
        commands = {}
        for cmd_pname in self.command_properties:
            cmd = self.wmeta.properties.get(cmd_pname, None)
            if cmd is not None:
                cmd = json.loads(cmd)
                cmd_name = (cmd["value"]).strip()
                if cmd_name:
                    commands[cmd_pname] = cmd
                else:
                    msg = "%s: invalid callback name for property '%s'."
                    logger.warning(msg, self.wmeta.identifier, cmd_pname)
        lines = []
        for cmd_pname, cmd in commands.items():
            callback = self._code_define_callback(cmd_pname, cmd)
            cmd_code = self._code_connect_command(cmd_pname, cmd, callback)
            if cmd_code:
                lines.extend(cmd_code)
        return lines

    def _code_define_callback_args(self, cmd_pname, cmd):
        cmdtype = cmd["cbtype"]
        args = None
        if cmdtype == CB_TYPES.WITH_WID:
            args = ("widget_id",)
        if cmdtype == CB_TYPES.SCALE:
            args = ("scale_value",)
        if cmdtype == CB_TYPES.SCROLLSET:
            args = ("first", "last")
        if cmdtype == CB_TYPES.SCROLL:
            args = ("mode", "value", "units")
        if cmdtype == CB_TYPES.ENTRY_VALIDATE:
            cb_args = cmd["args"]
            if cb_args:
                name_map = {
                    "%d": "d_action",
                    "%i": "i_index",
                    "%P": "p_entry_value",
                    "%s": "s_prev_value",
                    "%S": "s_new_value",
                    "%v": "v_validate",
                    "%V": "v_condition",
                    "%W": "w_entry_name",
                }
                args = []
                for key in cb_args.split():
                    args.append(name_map[key])

        return args

    def _code_define_callback(self, cmd_pname, cmd):
        cmdname = cmd["value"]
        cmdtype = cmd["cbtype"]
        args = self._code_define_callback_args(cmd_pname, cmd)
        wid = self.code_identifier()
        return self.builder.code_create_callback(wid, cmdname, cmdtype, args)

    def _code_connect_command(self, cmd_pname, cmd, cbname):
        target = self.code_identifier()
        args = cmd.get("args", "")
        args = args.split() if args else None
        lines = []
        cmdtype = cmd["cbtype"]
        if args is not None:
            if cmdtype == CB_TYPES.WITH_WID:
                wid = self.wmeta.identifier
                fdef = f'_wcmd = lambda wid="{wid}": {cbname}(wid)'
                cbname = "_wcmd"
                lines.append(fdef)
            if cmdtype == CB_TYPES.ENTRY_VALIDATE:
                original_cb = cbname
                tk_args = [f'"{a}"' for a in args]
                tk_args = ",".join(tk_args)
                fdef = f"_validatecmd = ({target}.register({original_cb}), {tk_args})"
                cbname = "_validatecmd"
                lines.append(fdef)
        line = f"{target}.configure({cmd_pname}={cbname})"
        lines.append(line)
        return lines

    def code_connect_bindings(self):
        lines = []
        target = self.code_identifier()
        for bind in self.wmeta.bindings:
            cb_name = self.builder.code_create_callback(
                target, bind.handler, CB_TYPES.BIND_EVENT
            )
            add_arg = "+" if bind.add else ""
            line = (
                f'{target}.bind("{bind.sequence}", {cb_name}, add="{add_arg}")'
            )
            lines.append(line)
        return lines

    def _code_process_layout_property(
        self, manager: str, pname: str, pvalue: str
    ) -> str:
        pvalue_str = str(pvalue)
        fvalue = (
            pvalue_str
            if pvalue_str.isnumeric() or isfloat(pvalue_str)
            else f'"{pvalue}"'
        )
        return fvalue

    def code_imports(self):
        # will return an iterable of (module, classname/function) to import
        # or None
        return None


#
# Base clases for some widgets
#
class EntryBaseBO(BuilderObject):
    """Base class for tk.Entry and ttk.Entry builder objects"""

    def _set_property(self, target_widget, pname, value):
        if pname == "text":
            wstate = str(target_widget["state"])
            if wstate != "normal":
                # change state temporarily
                target_widget["state"] = "normal"
            target_widget.delete("0", tk.END)
            target_widget.insert("0", value)
            target_widget["state"] = wstate
        else:
            super(EntryBaseBO, self)._set_property(target_widget, pname, value)

    #
    # Code generation methods
    #
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "text":
            fval = self.builder.code_translate_str(value)
            lines = [
                f"_text_ = {fval}",
                f'{targetid}.delete("0", "end")',
                f'{targetid}.insert("0", _text_)',
            ]
            if "state" in self.wmeta.properties:
                state_value = self.wmeta.properties["state"]
                if state_value != "normal":
                    lines.insert(1, f'{targetid}["state"] = "normal"')
                    line = f'{targetid}["state"] = "{state_value}"'
                    lines.append(line)
            code_bag[pname] = lines
        else:
            super(EntryBaseBO, self)._code_set_property(
                targetid, pname, value, code_bag
            )


class PanedWindowBO(BuilderObject):
    """Base class for tk.PanedWindow and ttk.Panedwindow builder objects"""

    class_ = None
    container = True
    properties = []
    ro_properties = ("orient",)

    def realize(self, parent):
        master = parent.get_child_master()
        args = self._get_init_args()
        if "orient" not in args:
            args["orient"] = "vertical"
        self.widget = self.class_(master, **args)
        return self.widget


#
# Base clases for some widget Helpers
#
class PanedWindowPaneBO(BuilderObject):
    class_ = None
    container = True
    properties = []
    layout_required = False
    allow_bindings = False

    def realize(self, parent):
        self.widget = parent.get_child_master()
        return self.widget

    def configure(self):
        pass

    def layout(self):
        pass

    def add_child(self, bobject):
        self.widget.add(bobject.widget, **self.wmeta.properties)

    #
    # code generation functions
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        return tuple()

    def code_configure(self, targetid=None):
        return tuple()

    def code_layout(self, targetid=None, parentid=None):
        return tuple()

    def code_child_add(self, childid):
        config = []
        masterid = self.code_child_master()
        for pname, val in self.wmeta.properties.items():
            line = f'{pname}="{val}"'
            config.append(line)
        kw = ""
        lines = []
        if config:
            kw = f""", {", ".join(config)}"""
        line = f"{masterid}.add({childid}{kw})"
        lines.append(line)
        return lines



# encoding: utf-8
import importlib
import logging
import tkinter
from pathlib import Path

from .component.builderobject import CLASS_MAP, BuilderObject
from .component.widgetmeta import WidgetMeta
from .stockimage import StockImage, StockImageException
from .component.uidefinition import UIDefinition
from .component.plugin_manager import PluginManager

logger = logging.getLogger(__name__)


#
# Builder class
#


class Builder(object):
    """Allows to build a tk interface from xml definition.

    Parameters
    ----------
    translator: callable
        function used to process translatable strings.
    """

    TK_VARIABLE_TYPES = ("string", "int", "boolean", "double")

    def __init__(self, translator=None):
        super(Builder, self).__init__()
        self.uidefinition = UIDefinition(translator=translator)
        self.root = None
        self.objects = {}
        self.tkvariables = {}
        self.translator = translator

    def add_resource_path(self, path):
        """Add additional path to the resources paths."""
        StockImage.add_resource_path(path)

    def add_resource_package(self, pkg: str):
        StockImage.add_resource_package(pkg)

    def get_image(self, path):
        """Return tk image corresponding to name which is taken form path."""
        image = ""
        name = Path(path).name
        if not StockImage.is_registered(name):
            StockImage.find_and_register(name)
        try:
            image = StockImage.get(name)
        except StockImageException:
            # TODO: notify something here.
            pass
        return image

    def get_iconbitmap(self, path):
        """Return path to use as iconbitmap property."""
        image = None
        name = Path(path).name
        if not StockImage.is_registered(name):
            StockImage.find_and_register(name)
        try:
            image = StockImage.as_iconbitmap(name)
        except StockImageException:
            # TODO: notify something here.
            pass
        return image

    def get_variable(self, varname):
        """Return a tk variable created with 'create_variable' method."""
        return self.tkvariables[varname]

    def import_variables(self, container, varnames=None):
        """Helper method to avoid call get_variable for every variable."""
        if varnames is None:
            for keyword in self.tkvariables:
                setattr(container, keyword, self.tkvariables[keyword])
        else:
            for keyword in varnames:
                if keyword in self.tkvariables:
                    setattr(container, keyword, self.tkvariables[keyword])

    def _process_variable_description(self, name_or_desc):
        vname = name_or_desc
        vtype = "string"  # default type if not defined
        if ":" in name_or_desc:
            vtype, vname = name_or_desc.split(":")
            #  Fix incorrect order bug #33
            if vtype not in self.TK_VARIABLE_TYPES:
                #  Swap order
                vtype, vname = vname, vtype
                if vtype not in self.TK_VARIABLE_TYPES:
                    msg = 'Undefined variable type in "{0}"'.format(vname)
                    raise Exception(msg)
        return (vname, vtype)

    def create_variable(self, varname, vtype=None):
        """Create a tk variable.
        If the variable was created previously return that instance.
        """
        vname, type_from_name = self._process_variable_description(varname)

        if vname in self.tkvariables:
            var = self.tkvariables[vname]
        else:
            if vtype is None:
                # get type from name
                if type_from_name == "int":
                    var = tkinter.IntVar()
                elif type_from_name == "boolean":
                    var = tkinter.BooleanVar()
                elif type_from_name == "double":
                    var = tkinter.DoubleVar()
                else:
                    var = tkinter.StringVar()
            else:
                var = vtype()

            self.tkvariables[vname] = var
        return var

    def add_from_file(self, file_or_filename):
        """Load ui definition from file."""
        self.uidefinition.load_file(file_or_filename)

    def add_from_string(self, strdata):
        self.uidefinition.load_from_string(strdata)

    def add_from_xmlnode(self, element):
        """Load ui definition from xml.etree.Element node."""
        self.uidefinition.add_xmlnode(element)

    def get_object(self, name, master=None):
        """Find and create the widget named name.
        Use master as parent. If widget was already created, return
        that instance."""
        widget = None
        if name in self.objects:
            widget = self.objects[name].widget
        else:
            wmeta = self.uidefinition.get_widget(name)
            if wmeta is not None:
                rmeta = WidgetMeta("root", "root")
                root = BuilderObject(self, rmeta)
                root.widget = master
                bobject = self._realize(root, wmeta)
                widget = bobject.widget
        if widget is None:
            msg = 'Widget "{0}" not defined.'.format(name)
            raise Exception(msg)
        return widget

    def _import_class(self, builder_id):
        plugin_managed = False
        for loader in PluginManager.builder_plugins():
            if loader.can_load(builder_id):
                _module = loader.get_module_for(builder_id)
                try:
                    importlib.import_module(_module)
                    plugin_managed = True
                    logger.debug("Module %s loaded.", _module)
                except (ModuleNotFoundError, ImportError) as e:
                    msg = "Failed to import module as fullname: %s"
                    logger.debug(msg, _module)
                    raise e

        # If no plugin, Try loading as old custom widget method.
        if not plugin_managed:
            _module = builder_id
            try:
                # Import module as full path
                importlib.import_module(_module)
                logger.debug("Module %s loaded.", _module)
            except (ModuleNotFoundError, ImportError) as e:
                # A single module can contain various widgets
                # try to import the first part of the path
                if "." in _module:
                    first, last = _module.rsplit(".", 1)
                    try:
                        importlib.import_module(first)
                        logger.debug("Module %s loaded.", first)
                    except (ModuleNotFoundError, ImportError):
                        importlib.import_module(last)
                        logger.debug("Module %s loaded.", last)
                else:
                    msg = "Failed to import module: %s"
                    logger.debug(msg, _module)
                    raise e

    def is_mapped(self, builder_uid):
        return builder_uid in CLASS_MAP

    def _get_builder_for(self, builder_uid):
        return CLASS_MAP[builder_uid].builder

    def _realize(self, master, wmeta):
        """Builds a widget from widget metadata using master as parent."""

        if not self.is_mapped(wmeta.classname):
            self._import_class(wmeta.classname)

        if self.is_mapped(wmeta.classname):
            bclass = self._get_builder_for(wmeta.classname)
            parent = bclass.factory(self, wmeta)
            self._pre_realize(parent)
            parent.realize(master)
            parent.configure()

            self.objects[wmeta.identifier] = parent

            for childmeta in self.uidefinition.widget_children(
                wmeta.identifier
            ):
                child = self._realize(parent, childmeta)
                parent.add_child(child)
            parent.configure_children()
            parent.layout()

            self._post_realize(parent)

            return parent
        else:
            msg = 'Class "{0}" not mapped'.format(wmeta.classname)
            raise Exception(msg)

    def _pre_realize(self, bobject):
        wmeta = bobject.wmeta
        cname = wmeta.classname
        wmeta.layout_required = bobject.layout_required
        has_layout = len(wmeta.layout_properties) > 1
        if wmeta.layout_required and not has_layout:
            logger.debug(
                "No layout information for: (%s, %s).",
                cname,
                wmeta.identifier,
            )

    def _post_realize(self, bobject):
        pass

    def connect_callbacks(self, callbacks_bag):
        """Connect callbacks specified in callbacks_bag with callbacks
        defined in the ui definition.
        Return a list with the name of the callbacks not connected.
        """
        notconnected = []
        for wname, builderobj in self.objects.items():
            missing = builderobj.connect_commands(callbacks_bag)
            if missing is not None:
                notconnected.extend(missing)
            missing = builderobj.connect_bindings(callbacks_bag)
            if missing is not None:
                notconnected.extend(missing)
        if notconnected:
            notconnected = list(set(notconnected))
            msg = "Missing callbacks for commands: %s"
            logger.warning(msg, notconnected)
            return notconnected
        else:
            return None

    def forget_unnamed(self):
        """Removes unnamed object from the objects attribute.

        User should calls this method at the end of the build process,
        after calling "connect_callbacks".
        """
        for key in list(self.objects.keys()):
            if not self.objects[key].wmeta.is_named:
                del self.objects[key]

    def code_create_variable(self, name_or_desc, value, vtype=None):
        raise NotImplementedError()

    def code_create_image(self, filename):
        raise NotImplementedError()

    def code_create_iconbitmap(self, filename):
        raise NotImplementedError()

    def code_classname_for(self, bobject):
        raise NotImplementedError()

    def code_create_callback(self, widgetid, cbname, cbtype, args=None):
        raise NotImplementedError()

    def code_translate_str(self, value: str) -> str:
        raise NotImplementedError()


# Load plugins on module init
PluginManager.load_plugins()



# encoding: utf-8
import logging
import tkinter as tk

from pygubu.i18n import _
from pygubu.api.v1 import BuilderObject, register_widget
from pygubu.component.builderobject import (
    CB_TYPES,
    EntryBaseBO,
    PanedWindowBO,
    PanedWindowPaneBO,
)

logger = logging.getLogger(__name__)

#
# tkinter widgets
#
_toplevel_87 = tuple()
if tk.TkVersion >= 8.7:
    _toplevel_87 = ("backgroundimage", "tile")


class TKToplevel(BuilderObject):
    class_ = tk.Toplevel
    container = True
    layout_required = False
    container_layout = True
    allowed_parents = ("root",)
    # maxchildren = 2  # A menu and a frame
    OPTIONS_STANDARD = (
        "borderwidth",
        "cursor",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "padx",
        "pady",
        "relief",
        "takefocus",
    )
    OPTIONS_SPECIFIC = (
        "background",
        "class_",
        "container",
        "height",
        "width",
    ) + _toplevel_87
    OPTIONS_CUSTOM = (
        "title",
        "geometry",
        "overrideredirect",
        "minsize",
        "maxsize",
        "resizable",
        "iconbitmap",
        "iconphoto",
    )
    ro_properties = ("container",)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    RESIZABLE = {
        "both": (True, True),
        "horizontally": (True, False),
        "vertically": (False, True),
        "none": (False, False),
    }

    def realize(self, parent):
        args = self._get_init_args()
        master = parent.get_child_master()
        if master is None and tk._default_root is None:
            self.widget = tk.Tk()
        else:
            self.widget = self.class_(master, **args)
        return self.widget

    # def layout(self, target=None):
    # we marked this widget as not allowed to edit layoutu
    #    pass

    def _set_property(self, target_widget, pname, value):
        method_props = ("geometry", "overrideredirect", "title")
        if pname in method_props:
            method = getattr(target_widget, pname)
            method(value)
        elif pname == "resizable" and value:
            target_widget.resizable(*self.RESIZABLE[value])
        elif pname == "maxsize":
            if "|" in value:
                w, h = value.split("|")
                target_widget.maxsize(w, h)
        elif pname == "minsize":
            if "|" in value:
                w, h = value.split("|")
                target_widget.minsize(w, h)
        elif pname == "iconphoto":
            icon = self.builder.get_image(value)
            target_widget.iconphoto(True, icon)
        elif pname == "iconbitmap":
            icon = self.builder.get_iconbitmap(value)
            target_widget.iconbitmap(icon)
        else:
            super(TKToplevel, self)._set_property(target_widget, pname, value)

    #
    # Code generation methods
    #
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname in ("geometry", "overrideredirect", "title"):
            line = f'{targetid}.{pname}("{value}")'
            code_bag[pname] = (line,)
        elif pname == "resizable":
            p1, p2 = self.RESIZABLE[value]
            line = "{0}.resizable({1}, {2})".format(targetid, p1, p2)
            code_bag[pname] = (line,)
        elif pname in ("maxsize", "minsize"):
            if "|" in value:
                w, h = value.split("|")
                line = "{0}.{1}({2}, {3})".format(targetid, pname, w, h)
                code_bag[pname] = (line,)
        elif pname == "iconbitmap":
            bitmap = self.builder.code_create_iconbitmap(value)
            line = f'{targetid}.iconbitmap("{bitmap}")'
            code_bag[pname] = (line,)
        elif pname == "iconphoto":
            image = self.builder.code_create_image(value)
            line = "{0}.iconphoto(True, {1})".format(targetid, image)
            code_bag[pname] = (line,)
        else:
            super(TKToplevel, self)._code_set_property(
                targetid, pname, value, code_bag
            )


register_widget(
    "tk.Toplevel", TKToplevel, "Toplevel", (_("Containers"), "tk", "ttk")
)


class TKFrame(BuilderObject):
    OPTIONS_STANDARD = (
        "borderwidth",
        "cursor",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "padx",
        "pady",
        "relief",
        "takefocus",
    )
    OPTIONS_SPECIFIC = (
        "background",
        "class_",
        "container",
        "height",
        "width",
    )
    class_ = tk.Frame
    container = True
    container_layout = True
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    ro_properties = ("class_", "container")


register_widget("tk.Frame", TKFrame, "Frame", (_("Containers"), "tk"))


class TKLabel(BuilderObject):
    OPTIONS_STANDARD = (
        "activebackground",
        "activeforeground",
        "anchor",
        "background",
        "bitmap",
        "borderwidth",
        "compound",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "height",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "image",
        "justify",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "text",
        "textvariable",
        "underline",
        "wraplength",
    )
    OPTIONS_SPECIFIC = ("height", "state", "width")
    class_ = tk.Label
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget("tk.Label", TKLabel, "Label", (_("Control & Display"), "tk"))


class TKLabelFrame(BuilderObject):
    class_ = tk.LabelFrame
    container = True
    container_layout = True
    OPTIONS_STANDARD = (
        "borderwidth",
        "cursor",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "text",
    )
    OPTIONS_SPECIFIC = (
        "background",
        "class_",
        "height",
        "labelanchor",
        "width",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    ro_properties = ("class_",)


register_widget(
    "tk.LabelFrame", TKLabelFrame, "LabelFrame", (_("Containers"), "tk")
)


class TKEntry(EntryBaseBO):
    OPTIONS_STANDARD = (
        "background",
        "borderwidth",
        "cursor",
        "exportselection",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "insertbackground",
        "insertborderwidth",
        "insertofftime",
        "insertontime",
        "insertwidth",
        "justify",
        "relief",
        "selectbackground",
        "selectborderwidth",
        "selectforeground",
        "takefocus",
        "textvariable",
        "xscrollcommand",
    )
    OPTIONS_SPECIFIC = (
        "disabledbackground",
        "disabledforeground",
        "invalidcommand",
        "readonlybackground",
        "show",
        "state",
        "validate",
        "validatecommand",
        "width",
    )
    OPTIONS_CUSTOM = ("text",)
    class_ = tk.Entry
    container = False
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = (
        "validatecommand",
        "invalidcommand",
        "xscrollcommand",
    )


register_widget("tk.Entry", TKEntry, "Entry", (_("Control & Display"), "tk"))


class TKButton(BuilderObject):
    class_ = tk.Button
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "activeforeground",
        "anchor",
        "background",
        "bitmap",
        "borderwidth",
        "compound",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "image",
        "justify",
        "padx",
        "pady",
        "relief",
        "repeatdelay",
        "repeatinterval",
        "takefocus",
        "text",
        "textvariable",
        "underline",
        "wraplength",
    )
    OPTIONS_SPECIFIC = (
        "command",
        "default",
        "height",
        "overrelief",
        "state",
        "width",
    )
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + BuilderObject.OPTIONS_CUSTOM
    )
    command_properties = ("command",)


register_widget("tk.Button", TKButton, "Button", (_("Control & Display"), "tk"))


class TKCheckbutton(BuilderObject):
    class_ = tk.Checkbutton
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "activeforeground",
        "anchor",
        "background",
        "bitmap",
        "borderwidth",
        "compound",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "image",
        "justify",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "text",
        "textvariable",
        "underline",
        "wraplength",
    )
    OPTIONS_SPECIFIC = (
        "command",
        "height",
        "indicatoron",
        "overrelief",
        "offrelief",
        "offvalue",
        "onvalue",
        "overrelief",
        "selectcolor",
        "selectimage",
        "state",
        "tristateimage",
        "tristatevalue",
        "variable",
        "width",
    )
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + BuilderObject.OPTIONS_CUSTOM
    )
    command_properties = ("command",)


register_widget(
    "tk.Checkbutton",
    TKCheckbutton,
    "Checkbutton",
    (_("Control & Display"), "tk"),
)


class TKRadiobutton(BuilderObject):
    class_ = tk.Radiobutton
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "activeforeground",
        "anchor",
        "background",
        "bitmap",
        "borderwidth",
        "compound",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "image",
        "justify",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "text",
        "textvariable",
        "underline",
        "wraplength",
    )
    OPTIONS_SPECIFIC = (
        "command",
        "height",
        "indicatoron",
        "overrelief",
        "offrelief",
        "overrelief",
        "selectcolor",
        "selectimage",
        "state",
        "tristateimage",
        "tristatevalue",
        "value",
        "variable",
        "width",
    )
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + BuilderObject.OPTIONS_CUSTOM
    )
    command_properties = ("command",)


register_widget(
    "tk.Radiobutton",
    TKRadiobutton,
    "Radiobutton",
    (_("Control & Display"), "tk"),
)


class TKListbox(BuilderObject):
    class_ = tk.Listbox
    container = False
    OPTIONS_STANDARD = (
        "background",
        "borderwidth",
        "cursor",
        "disabledforeground",
        "exportselection",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "justify",
        "relief",
        "selectbackground",
        "selectborderwidth",
        "selectforeground",
        "setgrid",
        "takefocus",
        "xscrollcommand",
        "yscrollcommand",
    )
    OPTIONS_SPECIFIC = (
        "activestyle",
        "height",
        "listvariable",
        "selectmode",
        "state",
        "width",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("xscrollcommand", "yscrollcommand")


register_widget(
    "tk.Listbox", TKListbox, "Listbox", (_("Control & Display"), "tk")
)


class TKText(BuilderObject):
    class_ = tk.Text
    container = False
    OPTIONS_STANDARD = (
        "background",
        "borderwidth",
        "cursor",
        "exportselection",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "insertbackground",
        "insertborderwidth",
        "insertofftime",
        "insertontime",
        "insertwidth",
        "padx",
        "pady",
        "relief",
        "selectbackground",
        "selectborderwidth",
        "selectforeground",
        "setgrid",
        "takefocus",
        "xscrollcommand",
        "yscrollcommand",
    )
    OPTIONS_SPECIFIC = (
        "autoseparators",
        "blockcursor",
        "endline",
        "height",
        "inactiveselectbackground",
        "insertunfocussed",
        "maxundo",
        "spacing1",
        "spacing2",
        "spacing3",
        "startline",
        "state",
        "tabs",
        "tabstyle",
        "undo",
        "width",
        "wrap",
    )
    OPTIONS_CUSTOM = ("text",)
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    command_properties = ("xscrollcommand", "yscrollcommand")

    def _set_property(self, target_widget, pname, value):
        if pname == "text":
            state = target_widget.cget("state")
            if state == tk.DISABLED:
                target_widget.configure(state=tk.NORMAL)
                target_widget.insert("0.0", value)
                target_widget.configure(state=tk.DISABLED)
            else:
                target_widget.insert("0.0", value)
        else:
            super(TKText, self)._set_property(target_widget, pname, value)

    #
    # Code generation methods
    #
    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "text":
            state_value = ""
            if "state" in self.wmeta.properties:
                state_value = self.wmeta.properties["state"]
            sval = self.builder.code_translate_str(value)
            lines = [
                f"_text_ = {sval}",
            ]
            if state_value == tk.DISABLED:
                lines.extend(
                    (
                        f'{targetid}.configure(state="normal")',
                        f'{targetid}.insert("0.0", _text_)',
                        f'{targetid}.configure(state="disabled")',
                    )
                )
            else:
                lines.append(f'{targetid}.insert("0.0", _text_)')
            code_bag[pname] = lines
        else:
            super(TKText, self)._code_set_property(
                targetid, pname, value, code_bag
            )


register_widget(
    "tk.Text", TKText, "Text", (_("Control & Display"), "tk", "ttk")
)


class TKPanedWindow(PanedWindowBO):
    class_ = tk.PanedWindow
    allowed_children = ("tk.PanedWindow.Pane",)
    OPTIONS_STANDARD = (
        "background",
        "borderwidth",
        "cursor",
        "orient",
        "relief",
    )
    OPTIONS_SPECIFIC = (
        "handlepad",
        "handlesize",
        "height",
        "opaqueresize",
        "proxybackground",
        "proxyborderwidth",
        "proxyrelief",
        "sashcursor",
        "sashpad",
        "sashrelief",
        "sashwidth",
        "showhandle",
        "width",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "tk.PanedWindow", TKPanedWindow, "PanedWindow", (_("Containers"), "tk")
)


class TKMenubutton(BuilderObject):
    class_ = tk.Menubutton
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "activeforeground",
        "anchor",
        "background",
        "bitmap",
        "borderwidth",
        "compound",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "image",
        "justify",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "text",
        "textvariable",
        "underline",
        "wraplength",
    )
    OPTIONS_SPECIFIC = (
        "direction",
        "height",
        "indicatoron",
        "state",
        "width",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    allowed_children = ("tk.Menu",)
    maxchildren = 1

    def add_child(self, bobject):
        self.widget.configure(menu=bobject.widget)

    def code_child_add(self, childid):
        lines = [f"{self.code_identifier()}.configure(menu={childid})"]
        return lines


register_widget(
    "tk.Menubutton",
    TKMenubutton,
    "Menubutton",
    (
        _("Menu"),
        _("Control & Display"),
        "tk",
    ),
)


class TKMessage(BuilderObject):
    class_ = tk.Message
    container = False
    OPTIONS_STANDARD = (
        "anchor",
        "background",
        "borderwidth",
        "cursor",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "padx",
        "pady",
        "relief",
        "takefocus",
        "text",
        "textvariable",
    )
    OPTIONS_SPECIFIC = ("aspect", "justify", "width")
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC


register_widget(
    "tk.Message", TKMessage, "Message", (_("Control & Display"), "tk", "ttk")
)


class TKScale(BuilderObject):
    class_ = tk.Scale
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "background",
        "borderwidth",
        "cursor",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "orient",
        "relief",
        "repeatdelay",
        "repeatinterval",
        "takefocus",
        "troughcolor",
    )
    OPTIONS_SPECIFIC = (
        "bigincrement",
        "command",
        "digits",
        "from_",
        "label",
        "length",
        "resolution",
        "showvalue",
        "sliderlength",
        "sliderrelief",
        "state",
        "tickinterval",
        "to",
        "variable",
        "width",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("command",)


register_widget("tk.Scale", TKScale, "Scale", (_("Control & Display"), "tk"))


class TKScrollbar(BuilderObject):
    class_ = tk.Scrollbar
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "background",
        "borderwidth",
        "cursor",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "jump",
        "orient",
        "relief",
        "repeatdelay",
        "repeatinterval",
        "takefocus",
        "troughcolor",
    )
    OPTIONS_SPECIFIC = (
        "activerelief",
        "command",
        "elementborderwidth",
        "width",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("command",)


register_widget(
    "tk.Scrollbar", TKScrollbar, "Scrollbar", (_("Control & Display"), "tk")
)


class TKSpinbox(BuilderObject):
    class_ = tk.Spinbox
    container = False
    OPTIONS_STANDARD = (
        "activebackground",
        "background",
        "borderwidth",
        "cursor",
        "exportselection",
        "font",
        "foreground",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "insertbackground",
        "insertborderwidth",
        "insertofftime",
        "insertontime",
        "insertwidth",
        "justify",
        "relief",
        "repeatdelay",
        "repeatinterval",
        "selectbackground",
        "selectborderwidth",
        "selectforeground",
        "takefocus",
        "textvariable",
        "xscrollcommand",
    )
    OPTIONS_SPECIFIC = (
        "buttonbackground",
        "buttoncursor",
        "buttondownrelief",
        "buttonuprelief",
        "command",
        "disabledbackground",
        "disabledforeground",
        "format",
        "from_",
        "invalidcommand",
        "increment",
        "readonlybackground",
        "state",
        "to",
        "validate",
        "validatecommand",
        "values",
        "width",
        "wrap",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = (
        "command",
        "invalidcommand",
        "validatecommand",
        "xscrollcommand",
    )

    def _set_property(self, target_widget, pname, value):
        # hack to configure 'from_' and 'to' and avoid exception
        if pname == "from_":
            from_ = float(value)
            to = float(self.wmeta.properties.get("to", 0))
            if from_ > to:
                to = from_ + 1
            target_widget.configure(from_=from_, to=to)
        else:
            super(TKSpinbox, self)._set_property(target_widget, pname, value)


register_widget(
    "tk.Spinbox", TKSpinbox, "Spinbox", (_("Control & Display"), "tk")
)


class TKCanvas(BuilderObject):
    class_ = tk.Canvas
    container = False
    OPTIONS_STANDARD = (
        "background",
        "borderwidth",
        "cursor",
        "highlightbackground",
        "highlightcolor",
        "highlightthickness",
        "insertbackground",
        "insertborderwidth",
        "insertofftime",
        "insertontime",
        "insertwidth",
        "relief",
        "selectbackground",
        "selectborderwidth",
        "selectforeground",
        "takefocus",
        "xscrollcommand",
        "yscrollcommand",
    )
    OPTIONS_SPECIFIC = (
        "closeenough",
        "confine",
        "height",
        "scrollregion",
        "state",
        "width",
        "xscrollincrement",
        "yscrollincrement",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("xscrollcommand", "yscrollcommand")


register_widget(
    "tk.Canvas", TKCanvas, "Canvas", (_("Control & Display"), "tk", "ttk")
)


class TKMenu(BuilderObject):
    layout_required = False
    allowed_parents = (
        "root",
        "tk.Menubutton",
        "ttk.Menubutton",
        "pygubu.builder.widgets.toplevelmenu",
    )
    allowed_children = (
        "tk.Menuitem.Submenu",
        "tk.Menuitem.Checkbutton",
        "tk.Menuitem.Command",
        "tk.Menuitem.Radiobutton",
        "tk.Menuitem.Separator",
    )
    class_ = tk.Menu
    container = True
    OPTIONS_STANDARD = (
        "activebackground",
        "activeborderwidth",
        "activeforeground",
        "background",
        "borderwidth",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "relief",
        "takefocus",
    )
    OPTIONS_SPECIFIC = (
        "postcommand",
        "selectcolor",
        "tearoff",
        "tearoffcommand",
        "title",
    )
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("postcommand", "tearoffcommand")
    allow_bindings = False

    def __init__(self, builder, wdescr):
        super().__init__(builder, wdescr)
        self._menuitems = None

    def layout(self, target=None):
        pass

    #
    # code generation functions
    #
    def add_menuitem(self, itemid, itemtype, properties):
        if self._menuitems is None:
            self._menuitems = []
        self._menuitems.append((itemtype, properties))

    def code_realize(self, boparent, code_identifier=None):
        start = -1
        tearoff_conf = self.wmeta.properties.get("tearoff", "true")
        if tearoff_conf in ("1", "true"):
            start = 0
        self.code_item_index = start
        return super(TKMenu, self).code_realize(boparent, code_identifier)

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname == "tearoffcommand":
            args = ("menu", "tearoff")
        return args

    def code_configure_children(self, targetid=None):
        if self._menuitems is None:
            return tuple()

        if targetid is None:
            targetid = self.code_identifier()
        lines = []
        for itemtype, kwp in self._menuitems:
            bag = []
            for pname, value in kwp.items():
                s = f"{pname}={value}"
                bag.append(s)
            props = ""
            if bag:
                props = ", " + ", ".join(bag)
            line = f'{targetid}.add("{itemtype}"{props})'
            lines.append(line)
        return lines


register_widget("tk.Menu", TKMenu, "Menu", (_("Menu"), "tk", "ttk"))


#
# Helpers for Standard tk widgets
#


class TKMenuitem(BuilderObject):
    class_ = None
    container = False
    itemtype = None
    layout_required = False
    OPTIONS_STANDARD = (
        "activebackground",
        "activeforeground",
        "background",
        "bitmap",
        "compound",
        "foreground",
        "state",
    )
    OPTIONS_SPECIFIC = (
        "accelerator",
        "columnbreak",
        "command",
        "font",
        "hidemargin",
        "image",
        "label",
        "underline",
    )
    properties = (
        OPTIONS_STANDARD + OPTIONS_SPECIFIC + BuilderObject.OPTIONS_CUSTOM
    )
    command_properties = ("command",)
    allow_bindings = False

    def realize(self, parent):
        self.widget = master = parent.get_child_master()
        itemproperties = dict(self.wmeta.properties)
        self._setup_item_properties(itemproperties)
        master.add(self.itemtype, **itemproperties)
        self._setup_item_index(parent)
        return self.widget

    def _setup_item_index(self, parent):
        master = parent.get_child_master()
        index = master.index(tk.END) or 0
        # TODO: index of items is shifted if tearoff is changed
        # for now check tearoff config and recalculate index.
        has_tearoff = True if master.type(0) == "tearoff" else False
        tearoff_conf = parent.wmeta.properties.get("tearoff", "1")
        offset = 0
        if has_tearoff and tearoff_conf in ("0", "false"):
            offset = 1
        self.__index = index - offset

    def _setup_item_properties(self, itemprops):
        for pname in itemprops:
            if pname == "variable":
                varname = itemprops[pname]
                itemprops[pname] = self.builder.create_variable(varname)
            if pname in ("image", "selectimage"):
                name = itemprops[pname]
                itemprops[pname] = self.builder.get_image(name)

    def configure(self):
        pass

    def layout(self, target=None):
        pass

    def _connect_command(self, cpname, callback):
        self.widget.entryconfigure(self.__index, command=callback)

    #
    # code generation functions
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        (
            code_bag,
            kwproperties,
            complex_properties,
        ) = self._code_process_properties(
            self.wmeta.properties, self._code_identifier, skip_commands=False
        )

        properties = {}
        for pname in kwproperties:
            properties[pname] = code_bag[pname]
        boparent.add_menuitem(self.wmeta.identifier, self.itemtype, properties)

        lines = []
        for pname in complex_properties:
            lines.extend(code_bag[pname])
        return lines

    def code_configure(self, targetid=None):
        return tuple()

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "command_id_arg":
            # pass, property for command configuration
            pass
        else:
            super(TKMenuitem, self)._code_set_property(
                targetid, pname, value, code_bag
            )

    def _pass_widgetid_to_callback(self):
        include_id = self.wmeta.properties.get(
            "command_id_arg", "false"
        ).lower()
        return include_id == "true"

    def _code_define_callback_args(self, cmd_pname, cmd):
        cmdtype = cmd["cbtype"]
        args = None
        if cmdtype == CB_TYPES.WITH_WID or self._pass_widgetid_to_callback():
            args = ("itemid",)
        return args

    def _code_define_callback(self, cmd_pname, cmd):
        usercb = super()._code_define_callback(cmd_pname, cmd)
        self._realcb = usercb
        args = self._code_define_callback_args(cmd_pname, cmd)
        if args is not None and "itemid" in args:
            usercb = f"{self.wmeta.identifier}_cmd"
        return usercb

    def _code_connect_item_command(self, cmd_pname, cmd, cbname):
        lines = []
        if cbname != self._realcb:
            cbname = self._realcb
            newcb = f"{self.wmeta.identifier}_cmd"
            wid = self.wmeta.identifier
            line = f'def {newcb}(itemid="{wid}"): {cbname}(itemid)'
            lines.append(line)
        return lines

    def _code_connect_command(self, cmd_pname, cmd, cbname):
        if self.itemtype != "submenu":
            return self._code_connect_item_command(cmd_pname, cmd, cbname)
        else:
            return super(TKMenuitem, self)._code_connect_command(
                cmd_pname, cmd, cbname
            )


class TKMenuitemSubmenu(TKMenuitem):
    itemtype = "submenu"
    allowed_parents = ("tk.Menu", "tk.Menuitem.Submenu")
    allowed_children = (
        "tk.Menuitem.Submenu",
        "tk.Menuitem.Checkbutton",
        "tk.Menuitem.Command",
        "tk.Menuitem.Radiobutton",
        "tk.Menuitem.Separator",
    )
    OPTIONS_STANDARD = (
        "activebackground",
        "activeborderwidth",
        "activeforeground",
        "background",
        "borderwidth",
        "bitmap",
        "compound",
        "cursor",
        "disabledforeground",
        "font",
        "foreground",
        "relief",
        "takefocus",
        "state",
    )
    OPTIONS_SPECIFIC = (
        "accelerator",
        "columnbreak",
        "hidemargin",
        "image",
        "label",
        "selectcolor",
        "tearoff",
        "tearoffcommand",
        "underline",
        "postcommand",
    )
    OPTIONS_CUSTOM = ("specialmenu",)
    properties = tuple(
        set(OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM)
    )
    # ro_properties = ('specialmenu', )
    command_properties = ("postcommand", "tearoffcommand")

    def __init__(self, builder, wdescr):
        super().__init__(builder, wdescr)
        self._menuitems = None

    def realize(self, parent):
        master = parent.get_child_master()
        self._setup_item_index(parent)

        menu_properties = dict(
            (k, v)
            for k, v in self.wmeta.properties.items()
            if k in TKMenu.properties or k == "specialmenu"
        )
        self._setup_item_properties(menu_properties)

        item_properties = dict(
            (k, v)
            for k, v in self.wmeta.properties.items()
            if k in TKMenuitem.properties
        )
        self._setup_item_properties(item_properties)

        self.widget = submenu = TKMenu.class_(master, **menu_properties)
        item_properties["menu"] = submenu
        master.add(tk.CASCADE, **item_properties)
        return self.widget

    def _setup_item_properties(self, itemprops):
        super(TKMenuitemSubmenu, self)._setup_item_properties(itemprops)
        pname = "specialmenu"
        if pname in itemprops:
            specialmenu = itemprops.pop(pname)
            itemprops["name"] = specialmenu

    def configure(self):
        pass

    def layout(self, target=None):
        pass

    def _connect_command(self, cpname, callback):
        # suported commands: tearoffcommand, postcommand
        kwargs = {cpname: callback}
        self.widget.configure(**kwargs)

    #
    # code generation functions
    #
    def add_menuitem(self, itemid, itemtype, properties):
        if self._menuitems is None:
            self._menuitems = []
        self._menuitems.append((itemtype, properties))

    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = code_identifier
        masterid = boparent.code_child_master()
        lines = []
        # menu properties
        menuprop = {}
        for pname, value in self.wmeta.properties.items():
            if pname in TKMenu.properties:
                menuprop[pname] = value
            if pname == "specialmenu":
                menuprop["name"] = value

        (
            code_bag,
            kw_properties,
            complex_properties,
        ) = self._code_process_properties(menuprop, self.code_identifier())
        for pname in complex_properties:
            lines.extend(code_bag[pname])

        mpbag = []
        for pname in kw_properties:
            line = "{0}={1}".format(pname, code_bag[pname])
            mpbag.append(line)
        mprops = ""
        if mpbag:
            mprops = ", " + ", ".join(mpbag)

        # item properties
        itemprop = {}
        for pname, value in self.wmeta.properties.items():
            if pname in TKMenuitem.properties:
                itemprop[pname] = value

        (
            code_bag,
            kw_properties,
            complex_properties,
        ) = self._code_process_properties(itemprop, self.code_identifier())
        for pname in complex_properties:
            lines.extend(code_bag[pname])

        pbag = []
        prop = "menu={0}".format(self.code_identifier())
        pbag.append(prop)
        for pname in kw_properties:
            line = "{0}={1}".format(pname, code_bag[pname])
            pbag.append(line)
        props = ""
        if pbag:
            props = ", {0}".format(", ".join(pbag))

        # creation
        line = "{0} = tk.Menu({1}{2})".format(
            self.code_identifier(), masterid, mprops
        )
        lines.append(line)
        line = "{0}.add(tk.CASCADE{1})".format(masterid, props)
        lines.append(line)

        return lines

    def code_configure(self, targetid=None):
        return tuple()

    def code_configure_children(self, targetid=None):
        if self._menuitems is None:
            return tuple()

        if targetid is None:
            targetid = self.code_identifier()
        lines = []
        for itemtype, kwp in self._menuitems:
            bag = []
            for pname, value in kwp.items():
                s = f"{pname}={value}"
                bag.append(s)
            props = ""
            if bag:
                props = ", " + ", ".join(bag)
            line = f'{targetid}.add("{itemtype}"{props})'
            lines.append(line)
        return lines

    def _code_define_callback_args(self, cmd_pname, cmd):
        args = None
        if cmd_pname == "tearoffcommand":
            args = ("menu", "tearoff")
        return args


register_widget(
    "tk.Menuitem.Submenu",
    TKMenuitemSubmenu,
    "Menuitem.Submenu",
    (_("Menu"), "tk", "ttk"),
)


class TKMenuitemCommand(TKMenuitem):
    allowed_parents = ("tk.Menu", "tk.Menuitem.Submenu")
    itemtype = tk.COMMAND


register_widget(
    "tk.Menuitem.Command",
    TKMenuitemCommand,
    "Menuitem.Command",
    (_("Menu"), "tk", "ttk"),
)


class TKMenuitemCheckbutton(TKMenuitem):
    allowed_parents = ("tk.Menu", "tk.Menuitem.Submenu")
    itemtype = tk.CHECKBUTTON
    OPTIONS_STANDARD = TKMenuitem.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TKMenuitem.OPTIONS_SPECIFIC + (
        "indicatoron",
        "onvalue",
        "offvalue",
        "selectcolor",
        "selectimage",
        "variable",
    )
    OPTIONS_CUSTOM = TKMenuitem.OPTIONS_CUSTOM
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM


register_widget(
    "tk.Menuitem.Checkbutton",
    TKMenuitemCheckbutton,
    "Menuitem.Checkbutton",
    (_("Menu"), "tk", "ttk"),
)


class TKMenuitemRadiobutton(TKMenuitem):
    allowed_parents = ("tk.Menu", "tk.Menuitem.Submenu")
    itemtype = tk.RADIOBUTTON
    OPTIONS_STANDARD = TKMenuitem.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TKMenuitem.OPTIONS_SPECIFIC + (
        "indicatoron",
        "selectcolor",
        "selectimage",
        "value",
        "variable",
    )
    OPTIONS_CUSTOM = TKMenuitem.OPTIONS_CUSTOM
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM


register_widget(
    "tk.Menuitem.Radiobutton",
    TKMenuitemRadiobutton,
    "Menuitem.Radiobutton",
    (_("Menu"), "tk", "ttk"),
)


class TKMenuitemSeparator(TKMenuitem):
    allowed_parents = ("tk.Menu", "tk.Menuitem.Submenu")
    itemtype = tk.SEPARATOR
    OPTIONS_STANDARD = ("background",)
    OPTIONS_SPECIFIC = tuple()
    OPTIONS_CUSTOM = tuple()
    properties = tuple()
    command_properties = tuple()


register_widget(
    "tk.Menuitem.Separator",
    TKMenuitemSeparator,
    "Menuitem.Separator",
    (_("Menu"), "tk", "ttk"),
)


class TKPanedWindowPane(PanedWindowPaneBO):
    class_ = None
    container = True
    allowed_parents = ("tk.PanedWindow",)
    maxchildren = 1
    OPTIONS_SPECIFIC = (
        "height",
        "hide",
        "minsize",
        "padx",
        "pady",
        "sticky",
        "stretch",
        "width",
    )
    properties = OPTIONS_SPECIFIC


register_widget(
    "tk.PanedWindow.Pane",
    TKPanedWindowPane,
    "PanedWindow.Pane",
    (_("Pygubu Helpers"), "tk"),
)


class TKLabelwidgetBO(BuilderObject):
    class_ = None
    container = True
    allowed_parents = ("tk.LabelFrame", "ttk.Labelframe")
    maxchildren = 1
    layout_required = False
    allow_bindings = False

    def realize(self, parent):
        self.widget = parent.get_child_master()
        return self.widget

    def add_child(self, bobject):
        self.widget.configure(labelwidget=bobject.widget)

    def layout(self, target=None):
        pass

    #
    # code generation functions
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        return tuple()

    def code_configure(self, targetid=None):
        return tuple()

    def code_layout(self, targetid=None, parentid=None):
        return tuple()

    def code_child_add(self, childid):
        line = "{0}.configure(labelwidget={1})"
        line = line.format(self.code_child_master(), childid)
        return (line,)


register_widget(
    "pygubu.builder.widgets.Labelwidget",
    TKLabelwidgetBO,
    "Labelwidget",
    (_("Pygubu Helpers"), "tk", "ttk"),
)


class ToplevelMenuHelperBO(BuilderObject):
    class_ = None
    container = True
    allowed_parents = ("tk.Toplevel",)
    maxchildren = 1
    layout_required = False
    allow_bindings = False

    def realize(self, parent):
        self.widget = parent.get_child_master()
        return self.widget

    def add_child(self, bobject):
        self.widget.configure(menu=bobject.widget)

    def layout(self, target=None):
        pass

    #
    # code generation functions
    #
    def code_realize(self, boparent, code_identifier=None):
        self._code_identifier = boparent.code_child_master()
        return tuple()

    def code_configure(self, targetid=None):
        return tuple()

    def code_layout(self, targetid=None, parentid=None):
        return tuple()

    def code_child_add(self, childid):
        line = "{0}.configure(menu={1})"
        line = line.format(self.code_child_master(), childid)
        return (line,)


register_widget(
    "pygubu.builder.widgets.toplevelmenu",
    ToplevelMenuHelperBO,
    "ToplevelMenu",
    (_("Menu"), _("Pygubu Helpers"), "tk", "ttk"),
)


class TKOptionMenu(BuilderObject):
    class_ = tk.OptionMenu
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = ("command", "variable", "value", "values")
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC
    command_properties = ("command",)
    ro_properties = ("variable", "value", "values")

    def realize(self, parent):
        args = self._get_init_args()
        master = parent.get_child_master()
        variable = args.pop("variable", None)
        value = args.pop("value", None)
        if variable is None:
            varname = "{0}__var".format(self.wmeta.identifier)
            variable = self.builder.create_variable(varname)
        if value is not None:
            variable.set(value)
        values = args.pop("values", "")
        if values is not None:
            values = values.split(",")

        class _cb_proxy(object):
            def __init__(self):
                super(_cb_proxy, self).__init__()
                self.callback = None

            def __call__(self, arg1):
                if self.callback is not None:
                    self.callback(arg1)

        cb_proxy = _cb_proxy()
        self.widget = self.class_(master, variable, *values, command=cb_proxy)
        self.widget._cb_proxy = cb_proxy
        return self.widget

    def _connect_command(self, cmd_pname, callback):
        if cmd_pname == "command":
            self.widget._cb_proxy.callback = callback

    #
    # Code generation methods
    #
    def code_realize(self, boparent, code_identifier=None):
        import json

        if code_identifier is not None:
            self._code_identifier = code_identifier
        lines = []
        master = boparent.code_child_master()
        init_args = self._code_get_init_args(self.code_identifier())
        command_arg = None
        variable_arg = None
        # value_arg = None

        # command property
        pname = "command"
        if pname in self.wmeta.properties:
            value = json.loads(self.wmeta.properties[pname])
            cmdname = value["value"]
            cmdtype = value["type"]
            args = ("option",)
            pvalue = self.builder.code_create_callback(
                self.code_identifier(), cmdname, cmdtype, args
            )
            command_arg = f"{pvalue}"

        # Value property
        # Not used in code generation.
        # Already setup in _code_get_init_args

        # Variable property
        pname = "variable"
        varname = "__tkvar"
        if pname in init_args:
            varname = init_args[pname]
        variable_arg = varname

        # values property
        pname = "values"
        om_values = []
        if pname in self.wmeta.properties:
            value = self.wmeta.properties["values"]
            om_values = value.split(",")
        line = f"__values = {om_values}"
        lines.append(line)
        s = f"{self.code_identifier()} = {self._code_class_name()}({master}, {variable_arg}, *__values, command={command_arg})"
        lines.append(s)
        return lines

    def code_configure(self, targetid=None):
        return []

    def code_connect_commands(self):
        return []


register_widget(
    "tk.OptionMenu",
    TKOptionMenu,
    "OptionMenu",
    (_("Control & Display"), "tk", "ttk"),
)



from abc import ABC, ABCMeta, abstractmethod
from typing import Optional


class PluginRegistry(ABCMeta):
    plugins = []

    def __init__(cls, name, bases, attrs):
        if name not in (
            "IPluginBase",
            "BuilderLoaderPlugin",
        ):
            PluginRegistry.plugins.append(cls)


class IDesignerPlugin(ABC):
    def get_preview_builder(self, builder_uid: str):
        """Return a BuilderObject subclass used to build a preview
        for the target builder_uid"""
        return None

    def get_toplevel_preview_for(
        self, builder_uid: str, widget_id: str, builder, top_master
    ):
        """Return a Toplevel with the target widget_id rendered inside."""
        return None

    def configure_for_preview(self, builder_uid: str, widget):
        """Make widget just display with minimal functionality."""
        pass


class IPluginBase(ABC, metaclass=PluginRegistry):
    @abstractmethod
    def do_activate(self) -> bool:
        "Initialize plugin and return if it is operational or not."
        ...

    def get_designer_plugin(self) -> Optional[IDesignerPlugin]:
        """Return class instance that implements IDesignerPlugin"""
        return None


class IBuilderLoaderPlugin(ABC):
    @abstractmethod
    def get_module_for(self, identifier: str) -> str:
        "Return module name for specified identifier."
        ...

    @abstractmethod
    def get_all_modules(self):
        "Return an iterable with module names of all builders."
        ...

    @abstractmethod
    def can_load(self, identifier: str) -> bool:
        "Return if this loader can manage specified identifier."
        ...


class BuilderLoaderPlugin(IBuilderLoaderPlugin, IPluginBase):
    pass



import tkinter as tk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _

from pygubu.plugins.tk.tkstdwidgets import TKFrame as TKFrameBO

from customtkinter import CTkBaseClass
from ..customtkinter import _designer_tab_label, _plugin_uid


class CTkBaseMixin:
    def _process_property_value(self, pname, value):
        if pname in ("hover", "dynamic_resizing"):
            return tk.getboolean(value)
        if pname in (
            "border_width",
            "from_",
            "to",
            "corner_radius",
            "button_corner_radius",
            "button_length",
            "number_of_steps",
        ):
            return float(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname in (
            "hover",
            "dynamic_resizing",
            "border_width",
            "from_",
            "to",
            "corner_radius",
            "button_corner_radius",
            "button_length",
            "number_of_steps",
        ):
            return super()._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)

    def code_imports(self):
        return [("customtkinter", self.class_.__name__)]


# I will register here all common properties used by customtkinter widgets
# so I don't have to repeat for each one (notice the .* at end of _builder_uid):
_builder_uid = f"{_plugin_uid}.*"
register_custom_property(
    _builder_uid,
    "bg_color",
    "colorentry",
    help=_("Color behind the widget if it has rounded corners."),
)
register_custom_property(_builder_uid, "border_color", "colorentry")
register_custom_property(_builder_uid, "border_width", "entry")
register_custom_property(_builder_uid, "button_color", "colorentry")
register_custom_property(_builder_uid, "button_hover_color", "colorentry")
register_custom_property(_builder_uid, "button_corner_radius", "entry")
register_custom_property(_builder_uid, "button_length", "entry")

register_custom_property(_builder_uid, "checkmark_color", "colorentry")
register_custom_property(_builder_uid, "command", "simplecommandentry")
register_custom_property(_builder_uid, "corner_radius", "entry")

register_custom_property(_builder_uid, "dropdown_color", "colorentry")
register_custom_property(_builder_uid, "dropdown_hover_color", "colorentry")
register_custom_property(_builder_uid, "dropdown_text_color", "colorentry")
register_custom_property(_builder_uid, "dropdown_text_font", "fontentry")
register_custom_property(
    _builder_uid,
    "dynamic_resizing",
    "choice",
    values=("", "True", "False"),
    state="readonly",
)

register_custom_property(
    _builder_uid, "fg_color", "colorentry", help=_("Main color of the widget.")
)

register_custom_property(_builder_uid, "height", "dimensionentry")
register_custom_property(
    _builder_uid,
    "hover",
    "choice",
    values=("", "True", "False"),
    state="readonly",
)
register_custom_property(_builder_uid, "hover_color", "colorentry")

register_custom_property(_builder_uid, "number_of_steps", "entry")

register_custom_property(_builder_uid, "placeholder_text", "entry")
register_custom_property(_builder_uid, "placeholder_text_color", "colorentry")
register_custom_property(_builder_uid, "progress_color", "colorentry")

register_custom_property(
    _builder_uid,
    "state",
    "choice",
    values=("", "normal", "active", "disabled"),
    state="readonly",
)

register_custom_property(_builder_uid, "text", "text")
register_custom_property(_builder_uid, "text_color", "colorentry")
register_custom_property(_builder_uid, "text_color_disabled", "colorentry")
register_custom_property(_builder_uid, "text_font", "fontentry")

register_custom_property(_builder_uid, "values", "entry")
register_custom_property(_builder_uid, "variable", "tkvarentry")

register_custom_property(_builder_uid, "width", "dimensionentry")

register_custom_property(
    _builder_uid,
    "appearance_mode",
    "choice",
    values=("", "dark", "light"),
    state="readonly",
)

register_custom_property(
    _builder_uid,
    "color_theme",
    "choice",
    values=("", "blue", "green", "dark-blue", "sweetkind"),
    state="readonly",
    help=_("Default color theme."),
)



from pygubu.api.v1 import BuilderObject


class AutocompleteBaseBO(BuilderObject):
    init_completevalues = True

    def _get_init_args(self):
        args = super(AutocompleteBaseBO, self)._get_init_args()
        if self.init_completevalues:
            prop = "completevalues"
            if prop not in self.wmeta.properties:
                args[prop] = ""
        return args

    def _process_property_value(self, pname, value):
        final_value = value
        if pname == "completevalues":
            final_value = value.split()
        else:
            final_value = super(
                AutocompleteBaseBO, self
            )._process_property_value(pname, value)
        return final_value

    def _code_process_property_value(self, targetid, pname, value):
        if pname == "completevalues":
            cvalues = None
            try:
                cvalues = f"{value.split()}"
            except Exception:
                pass
            return cvalues
        return super()._code_process_property_value(targetid, pname, value)


class CallbakInitArgMixin:
    """Some widgtes of the ttkwidget set, have a callback argument in the
    constructor and the widget does not allow to use configure method
    to modify it.
    So create a mixin class to configure the callback with a trick.
    """

    def _get_init_args(self):
        args = super(CallbakInitArgMixin, self)._get_init_args()

        class CBProxy:
            def __init__(self):
                self.real_cb = None

            def __call__(self, family):
                if self.real_cb is not None:
                    self.real_cb(family)

        args["callback"] = cb = CBProxy()
        self._cb_proxy = cb
        return args

    def _connect_command(self, cmd_pname, callback):
        # continue the trick for setting the callback init argument
        if cmd_pname == "callback":
            self._cb_proxy.real_cb = callback
        else:
            super(CallbakInitArgMixin, self)._connect_command(
                cmd_pname, callback
            )

    def _code_get_init_args(self, code_identifier):
        args = super(CallbakInitArgMixin, self)._code_get_init_args(
            code_identifier
        )
        pname = "callback"
        if pname in self.wmeta.properties:
            pvalue = self.wmeta.properties[pname]
            args["callback"] = self._code_process_property_value(
                code_identifier, pname, pvalue
            )
        return args

    def _code_connect_command(self, cmd_pname, cmd, cbname):
        if cmd_pname == "callback":
            return []
        else:
            return super(CallbakInitArgMixin, self)._code_connect_command(
                cmd_pname, cmd, cbname
            )



from pygubu.api.v1 import IDesignerPlugin
from .toplevelframe import ToplevelFramePreviewBO


class PygubuDesignerPlugin(IDesignerPlugin):
    def get_preview_builder(self, builder_uid: str):
        if builder_uid in ("tk.Toplevel", "pygubu.builder.widgets.dialog"):
            return ToplevelFramePreviewBO
        return None

    def get_toplevel_preview_for(
        self, builder_uid: str, widget_id: str, builder, top_master
    ):
        top = None
        if builder_uid == "tk.Toplevel":
            top = builder.get_object(widget_id, top_master)
        elif builder_uid == "pygubu.builder.widgets.dialog":
            dialog = builder.get_object(widget_id, top_master)
            dialog.run()
            top = dialog.toplevel
        return top

    def configure_for_preview(self, builder_uid: str, widget):
        """Make widget just display with minimal functionality."""
        pass



# encoding: utf-8
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKCombobox
from pygubu.i18n import _
from pygubu.widgets.combobox import Combobox


class ComboboxBuilder(TTKCombobox):
    OPTIONS_SPECIFIC = tuple(
        set(TTKCombobox.OPTIONS_SPECIFIC) - set(("state",))
    )
    OPTIONS_CUSTOM = TTKCombobox.OPTIONS_CUSTOM + ("keyvariable",)
    properties = (
        TTKCombobox.OPTIONS_STANDARD
        + TTKCombobox.OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    tkvar_properties = TTKCombobox.tkvar_properties + ("keyvariable",)
    class_ = Combobox


_builder_id = "pygubu.builder.widgets.combobox"
register_widget(
    _builder_id, ComboboxBuilder, "Combobox", ("ttk", _("Pygubu Widgets"))
)

_help = "Tk variable associated to the key value."
register_custom_property(_builder_id, "keyvariable", "tkvarentry", help=_help)
_help = "Combobox state."
register_custom_property(
    _builder_id,
    "state",
    "choice",
    values=("", "normal", "disabled"),
    state="readonly",
    help=_help,
)
